/*
 *  @file   FrameQBufMgr_ShMem.c
 *
 *  @brief  Implementation of FrameQBufMgr ShMem module on RTOS
 *          side(DSP/BIOS).
 *
 *
 *  ============================================================================
 *
 *  Copyright (c) 2008-2009, Texas Instruments Incorporated
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  
 *  *  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  
 *  *  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  
 *  *  Neither the name of Texas Instruments Incorporated nor the names of
 *     its contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *  
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 *  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 *  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 *  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 *  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *  Contact information for paper mail:
 *  Texas Instruments
 *  Post Office Box 655303
 *  Dallas, Texas 75265
 *  Contact information: 
 *  http://www-k.ext.ti.com/sc/technical-support/product-information-centers.htm?
 *  DCMP=TIHomeTracking&HQS=Other+OT+home_d_contact
 *  ============================================================================
 *  
 */



/*-------------------------    XDC  specific includes ------------------------*/
#include <xdc/std.h>
#include <xdc/runtime/Memory.h>
#include <xdc/runtime/System.h>
#include <xdc/runtime/Startup.h>
#include <xdc/runtime/Assert.h>
#include <xdc/runtime/Error.h>
#include <xdc/runtime/IGateProvider.h>
#include <xdc/runtime/Gate.h>
#include <ti/sysbios/gates/GateSwi.h>

/*-------------------------    Generic includes  -----------------------------*/
#include <string.h>

/*-------------------------    BIOS includes    ------------------------------*/
#include <ti/sysbios/gates/GateMutex.h>

/*-------------------------    IPC module specific includes ------------------*/
#include <ti/sdo/ipc/Ipc.h>

#include <ti/syslink/inc/SyslinkMemMgr_errBase.h>
#include <ti/syslink/inc/ClientNotifyMgr.h>

#include <ti/syslink/ipc/rtos/interfaces/IFrameQBufMgr.h>
#ifdef xdc_target__isaCompatible_64P
#include <ti/sysbios/family/c64p/Cache.h>
#else
#include <ti/sysbios/hal/Cache.h>
#endif
#include "package/internal/FrameQBufMgr_ShMem.xdc.h"

#include <ti/syslink/ipc/rtos/interfaces/_SysLink.h>
#include <ti/sdo/utils/_NameServer.h>
#include <ti/sdo/ipc/_ListMP.h>
#include <ti/sdo/ipc/_SharedRegion.h>
#include <ti/sdo/ipc/_GateMP.h>
#include <ti/sdo/utils/_MultiProc.h>
#include <ti/sdo/ipc/heaps/_HeapMemMP.h>


#include <ti/syslink/ipc/rtos/_FrameQBufMgr.h>
#include <ti/syslink/ipc/rtos/shMem/_FrameQBufMgr_ShMem.h>
#include <ti/syslink/ipc/rtos/_FrameQ.h>
#include <ti/syslink/ipc/rtos/syslinkMemMgr/SyslinkMemMgr.h>

#define ROUND_UP(a, b) (SizeT)((((UInt32) a) + ((b) - 1)) & ~((b) - 1))

#define FRAMEQBUFMGR_NAMESERVERNAME          "FrameQBufMgr"

/*TODO: Move these defines to xdc interface file */
#define FRAMEQBUFMGR_INSTNO_BITS             8u
#define FRAMEQBUFMGR_INSTNO_MASK             0xFF
#define FRAMEQBUFMGR_INSTNO_BITOFFSET        0u
#define FRAMEQBUFMGR_CREATORPROCID_BITS      8u
#define FRAMEQBUFMGR_CREATORPROCID_MASK      0xFF
#define FRAMEQBUFMGR_CREATORPROCID_BITOFFSET 8u


extern Ptr gFrameQBufMgr_nsHandle;

/*=============================================================================
 * Forward declarations of internal Functions
 *=============================================================================
 */
Void FrameQBufMgr_Shmem_updateCacheFlags(FrameQBufMgr_ShMem_Object *obj,
                                         UInt32 cacheFlags);

static inline
Int32
_FrameQBufMgr_ShMem_create(FrameQBufMgr_ShMem_Object       * obj,
                           const ti_syslink_ipc_rtos_shMem_FrameQBufMgr_ShMem_Params * params,
                           Error_Block *eb);

static inline Int32
_FrameQBufMgr_ShMem_open(FrameQBufMgr_ShMem_Object       * obj,
                         const ti_syslink_ipc_rtos_shMem_FrameQBufMgr_ShMem_Params * params,
                         Error_Block *eb);

static inline
Int32
_FrameQBufMgr_ShMem_openHdrBufMemMgr(
                                FrameQBufMgr_ShMem_Object *obj,
                                FrameQBufMgr_NameServerEntry *pnameserverEntry,
                                Error_Block *eb);

static inline
Int32
_FrameQBufMgr_ShMem_openFrmBufMemMgr(
                                FrameQBufMgr_ShMem_Object *obj,
                                FrameQBufMgr_NameServerEntry *pnameserverEntry,
                                Error_Block *eb);

/*
 *  ======== FrameQBufMgr_ShMem_getParams ========
 *  API to get the module wide config params.
 */
Int32 FrameQBufMgr_ShMem_getConfig (FrameQBufMgr_ShMem_Config* cfg)
{
    Int32 status = FrameQBufMgr_S_SUCCESS;

    Assert_isTrue ((NULL != cfg), FrameQBufMgr_ShMem_A_invalidParams);

    if (NULL != cfg) {
        /* (If setup has not yet been called) */
        if (FrameQBufMgr_ShMem_module->refCount == 0) {
            /* Return the default static config values */
            cfg->eventNo        = FrameQBufMgr_ShMem_NOTIFY_RESERVED_EVENTNO;
            cfg->useNameServer  = TRUE;
            cfg->maxInstances   = FrameQBufMgr_MAXINSTANCES;
        }
        else {
            /* Return the currently configured parameters */
            memcpy (cfg,
                    &FrameQBufMgr_ShMem_module->cfg,
                    sizeof (FrameQBufMgr_ShMem_Config));
        }

    }
    else {
        status = FrameQBufMgr_E_INVALIDARG;
    }

    return (status);
}

/*
 *  ======== FrameQBufMgr_ShMem_sharedMemReq ========
 *  API to find shared memory requirements for the instance.
 */
UInt32
FrameQBufMgr_ShMem_sharedMemReq(
                                const FrameQBufMgr_ShMem_Params *params,
                                UInt32                    *frmHdrBufChunkSize,
                                UInt32                    *frmBufChunkSize)
{
    UInt32                  totalSize        = 0;
    UInt32                  bufSize          = 0;
    UInt32                  frmBufSizesInFrm = 0;
    ClientNotifyMgr_Params  clientMgrParams;
    GateMP_Params           gateParams;
    UInt32                  i;
    UInt32                  j;
    SizeT                   minAlign;
    SizeT                   hdrBufMinAlign;
    SizeT                   frmBufMinAlign;
    UInt16                  regionId;
    UInt16                  hdrBuf_regionId;
    UInt16                  frmBuf_regionId;

    Assert_isTrue ((NULL != params), FrameQBufMgr_A_invalidParams);

    if (NULL != params) {
        Assert_isTrue((params->numFreeFramePools <= FrameQBufMgr_ShMem_MAX_POOLS),
                      FrameQBufMgr_ShMem_A_invalidParams);

        if (params->sharedAddr) {
            regionId = SharedRegion_getId(params->sharedAddr);
        }
        else {
            regionId = params->regionId;
        }
        Assert_isTrue(regionId != SharedRegion_INVALIDREGIONID,
                      Ipc_A_internal);

        minAlign = Memory_getMaxDefaultTypeAlign();
        if (SharedRegion_getCacheLineSize(regionId) > minAlign) {
            minAlign = SharedRegion_getCacheLineSize(regionId);
        }

        /* Header buffers */
        if (params->sharedAddrHdrBuf) {
            hdrBuf_regionId = SharedRegion_getId(params->sharedAddrHdrBuf);
        }
        else {
            hdrBuf_regionId = params->hdrBuf_regionId;
        }
        Assert_isTrue(hdrBuf_regionId != SharedRegion_INVALIDREGIONID,
                      Ipc_A_internal);

        hdrBufMinAlign = Memory_getMaxDefaultTypeAlign();
        if (SharedRegion_getCacheLineSize(hdrBuf_regionId) > hdrBufMinAlign) {
            hdrBufMinAlign = SharedRegion_getCacheLineSize(hdrBuf_regionId);
        }

        /* FrameQ buffers */
        if (params->sharedAddrFrmBuf) {
            frmBuf_regionId = SharedRegion_getId(params->sharedAddrFrmBuf);
        }
        else {
            frmBuf_regionId = params->frmBuf_regionId;
        }
        Assert_isTrue(frmBuf_regionId != SharedRegion_INVALIDREGIONID,
                      Ipc_A_internal);

        frmBufMinAlign = Memory_getMaxDefaultTypeAlign();
        if (SharedRegion_getCacheLineSize(frmBuf_regionId) > frmBufMinAlign) {
            frmBufMinAlign = SharedRegion_getCacheLineSize(frmBuf_regionId);
        }


        totalSize = ROUND_UP(sizeof (FrameQBufMgr_NameServerEntry), minAlign);

        /* Calculate the memory required for the instance
         * shared control structure
         */
        totalSize += ROUND_UP(sizeof (FrameQBufMgr_ShMem_Attrs), minAlign);

        totalSize += ROUND_UP(sizeof (FrameQBufMgr_ShMem_FreeFramePoolObj),
                               minAlign) * params->numFreeFramePools;
        if (FrameQBufMgr_ShMem_usedefaultgate == FALSE) {
            if(params->gate == NULL) {
                GateMP_Params_init(&gateParams);
                gateParams.regionId = regionId;
                gateParams.remoteProtect =
                                  (GateMP_RemoteProtect) params->remoteProtect;

                gateParams.localProtect  =
                                      (GateMP_LocalProtect)params->localProtect;
                totalSize += ROUND_UP(GateMP_sharedMemReq(&gateParams),
                                       minAlign);/* for self */
                gateParams.localProtect  =
                             (GateMP_LocalProtect)params->localProtect;
                totalSize += ROUND_UP(GateMP_sharedMemReq(&gateParams),
                                       minAlign);/* For ClientNotifyMgr */
            }
        }
        /* Calculate the memory required for the instance's
         * ClientNotifyMgr handle
         */
        ClientNotifyMgr_Params_init (&clientMgrParams);
        clientMgrParams.regionId = regionId;
        clientMgrParams.numNotifyEntries = params->numNotifyEntries;
        clientMgrParams.numSubNotifyEntries = params->numFreeFramePools;
        totalSize +=  ROUND_UP(ClientNotifyMgr_sharedMemReq (
                                    (ClientNotifyMgr_Params*)&clientMgrParams ),
                                   minAlign);

        /* Calculate buffer Chunk needed for frame header buffers */

        for (i = 0; i < params->numFreeFramePools; i++) {
            bufSize += (   (    params->numFreeFramesInPool[i]
                              + params->numFreeHeaderBufsInPool[i])
                        * (  ROUND_UP(params->frameHeaderBufSize[i],
                             hdrBufMinAlign) ));
        }
        if (frmHdrBufChunkSize != NULL) {
            *frmHdrBufChunkSize = bufSize;
        }


        /* Calculate buffer Chunk needed for frame buffers */
        /* Reset bufSize */
        bufSize = 0;
        for (i = 0; i < params->numFreeFramePools; i++) {
            for (j = 0; j < params->numFrameBufsInFrame[i]; j++ ) {
                if (params->bufInterfaceType ==
                    FrameQBufMgr_ShMem_BUFINTERFACE_SHAREDMEM) {
                frmBufSizesInFrm +=
                       ROUND_UP((UInt32)params->frameBufParams[i][j].size,
                             frmBufMinAlign);
            }
                if (params->bufInterfaceType ==
                    FrameQBufMgr_ShMem_BUFINTERFACE_TILERMEM) {
                        frmBufSizesInFrm = 0;
                      /* TBD */
                }
            }
            bufSize += (frmBufSizesInFrm * params->numFreeFramesInPool[i]);
            frmBufSizesInFrm = 0;
        }

        if (frmBufChunkSize != NULL) {
            *frmBufChunkSize = bufSize;
        }

    }
    return (totalSize);
}

/*
 *  ======== FrameQBufMgr_ShMem_getSharedParams ========
 */
static Void FrameQBufMgr_ShMem_getSharedParams(FrameQBufMgr_ShMem_Params *params,
    const ti_syslink_ipc_rtos_shMem_FrameQBufMgr_ShMem_Params *lShMemParams)
{

    UInt32 i;

    /*Copy contens to shim params */
    memcpy(&(params->commonCreateParams),
           &(lShMemParams->commonCreateParams),
            sizeof (FrameQBufMgr_ShMem_CreateParams));


    params->gate                    = (GateMP_Handle)lShMemParams->gate;
    params->localProtect = (GateMP_LocalProtect)
                                                   lShMemParams->localProtect;
    params->remoteProtect = (GateMP_RemoteProtect)
                                                   lShMemParams->remoteProtect;

    params->sharedAddr              = lShMemParams->sharedAddr;
    params->sharedAddrSize          = lShMemParams->sharedAddrSize;
    params->sharedAddrHdrBuf        = lShMemParams->sharedAddrHdrBuf ;
    params->sharedAddrHdrBufSize    = lShMemParams->sharedAddrHdrBufSize;
    params->sharedAddrFrmBuf        = lShMemParams->sharedAddrFrmBuf;
    params->sharedAddrFrmBufSize    = lShMemParams->sharedAddrFrmBufSize;
    params->numFreeFramePools       = lShMemParams->numFreeFramePools;
    params->regionId                = lShMemParams->regionId;
    params->hdrBuf_regionId         = lShMemParams->hdrBuf_regionId;
    params->frmBuf_regionId         = lShMemParams->frmBuf_regionId;
    params->cpuAccessFlags          = lShMemParams->cpuAccessFlags ;
    params->numNotifyEntries        = lShMemParams->numNotifyEntries ;
    params->headerInterfaceType     = lShMemParams->headerInterfaceType;
    params->bufInterfaceType        = lShMemParams->bufInterfaceType;

    for (i = 0; i < FrameQBufMgr_MAX_POOLS; i++) {
        params->numFreeFramesInPool[i]     =
                                           lShMemParams->numFreeFramesInPool[i];
        params->numFreeHeaderBufsInPool[i] =
                                       lShMemParams->numFreeHeaderBufsInPool[i];
        params->frameHeaderBufSize[i]      =
                                            lShMemParams->frameHeaderBufSize[i];
        params->numFrameBufsInFrame[i]     =
                                           lShMemParams->numFrameBufsInFrame[i];

        params->frameBufParams[i] = (FrameQBufMgr_FrameBufParams*)
                                               lShMemParams->frameBufParams[i];
    }
}

/*
 *  ======== FrameQBufMgr_ShMem_Params_init ========
 *  Shim for function to get default initialization param for the FrameQ shMem
 *  instance.
 */
void
FrameQBufMgr_ShMem_Params_init(FrameQBufMgr_ShMem_Params * params)
{
    ti_syslink_ipc_rtos_shMem_FrameQBufMgr_ShMem_Params lShMemParams;
    UInt32 i;

    ti_syslink_ipc_rtos_shMem_FrameQBufMgr_ShMem_Params_init(&lShMemParams);

    /*Copy contens to shim params */
    memcpy(&(params->commonCreateParams),
           &(lShMemParams.commonCreateParams),
            sizeof (FrameQBufMgr_ShMem_CreateParams));

    for (i = 0; i < FrameQBufMgr_ShMem_MAX_POOLS; i++) {
        params->numFreeFramesInPool[i]     = 0;
        params->numFreeHeaderBufsInPool[i] = 0;
        params->frameHeaderBufSize[i]      = 0;
        params->numFrameBufsInFrame[i]     = 0;

        params->frameBufParams[i] = NULL;

    }

    params->gate                    = (GateMP_Handle)lShMemParams.gate;
    params->sharedAddr              = lShMemParams.sharedAddr;
    params->sharedAddrSize          = lShMemParams.sharedAddrSize;
    params->sharedAddrHdrBuf        = lShMemParams.sharedAddrHdrBuf ;
    params->sharedAddrHdrBufSize    = lShMemParams.sharedAddrHdrBufSize;
    params->sharedAddrFrmBuf        = lShMemParams.sharedAddrFrmBuf;
    params->sharedAddrFrmBufSize    = lShMemParams.sharedAddrFrmBufSize;
    params->numFreeFramePools       = lShMemParams.numFreeFramePools;

    /* Pass HeapMemMp names to used  for MemMgrs */
    params->regionId                = lShMemParams.regionId;
    params->hdrBuf_regionId         = lShMemParams.hdrBuf_regionId;
    params->frmBuf_regionId         = lShMemParams.frmBuf_regionId;
    params->cpuAccessFlags          = lShMemParams.cpuAccessFlags ;
    params->numNotifyEntries        = lShMemParams.numNotifyEntries;
}

/* -------- per-instance functions -------- */

/*
 *  ======== FrameQBufMgr_Instance_init ========
 *  API to create/open the instance.Called internally by
 *  the FrameQBufMgr_create call.
 */
Int
FrameQBufMgr_ShMem_Instance_init(FrameQBufMgr_ShMem_Object       * obj,
                                 const ti_syslink_ipc_rtos_shMem_FrameQBufMgr_ShMem_Params * params,
                                 Error_Block *eb)
{
    Int32   status  = FrameQBufMgr_S_SUCCESS;

    Assert_isTrue ((NULL != obj), FrameQBufMgr_ShMem_A_invalidParams);

    Assert_isTrue ((NULL != params), FrameQBufMgr_ShMem_A_invalidParams);

    if (params->commonCreateParams.openFlag == TRUE) {
        /* Call internal open call */
        status = _FrameQBufMgr_ShMem_open(obj, params, eb);

    }
    else {
        /* Call internal create call */
        status = _FrameQBufMgr_ShMem_create(obj, params, eb);
    }

    return (status);
}

/*
 *  ======== FrameQBufMgr_Instance_finalize ========
 *  API to finalize the created instance.
 */
 Void
 FrameQBufMgr_ShMem_Instance_finalize( FrameQBufMgr_ShMem_Object* obj ,
                                  int status )
 {
    volatile FrameQBufMgr_ShMem_Attrs  *pAttrs          = NULL;
    FrameQBufMgr_ShMem_FrameBufInfo              *frameBufInfo  = NULL;
    volatile FrameQBufMgr_ShMem_FreeFramePoolObj *freeFramePoolObj;
    UInt32                        key;
    UInt16                        instId;
    UInt16                        instNo;
    UInt16                        creProcId;
    UInt32                        i;
    UInt32                        j;
    UInt32                        k;
    FrameQBufMgr_ShMem_Frame      frame;
    Ptr                        portableAddr;
    SyslinkMemMgr_freeParams      freeParams;

    Assert_isTrue ((NULL != obj), FrameQBufMgr_A_invalidParams);

    pAttrs = obj->attrs;

    switch (obj->objType) {

    case FrameQBufMgr_ShMem_DYNAMIC_OPEN:
    {
        key = GateMP_enter ((GateMP_Handle)obj->gate);

        if (obj->ctrlStructCacheFlag == TRUE) {
            Cache_inv ((Ptr)pAttrs,
                       sizeof (FrameQBufMgr_ShMem_Attrs),
                       Cache_Type_ALL,
                       TRUE);
        }

        GateMP_leave ((GateMP_Handle)obj->gate, key);

        /*Free the allocated memory for obj->freeFramePool pointers */
        if (obj->freeFramePool != NULL) {
            Memory_free (NULL,
                         obj->freeFramePool,
                         sizeof(UInt32)* pAttrs->numFreeFramePools);

        }

        /*Unregister the notification with the clientNotifyMgr */
        if (obj->isRegistered == TRUE) {
            ClientNotifyMgr_unregisterClient ((ClientNotifyMgr_Handle)obj->clientNotifyMgrHandle,
                                              obj->notifyId);
            obj->isRegistered = FALSE;
        }

        if (obj->clientNotifyMgrHandle != NULL) {
            ClientNotifyMgr_close ((ClientNotifyMgr_Handle *)&(obj->clientNotifyMgrHandle));
        }

        if (   (obj->creCliGate == TRUE)
            && (obj->clientNotifyMgrGate != NULL) ) {
            /* Close GateMP instance of ClientNotifyMgr.*/
            GateMP_close ((GateMP_Handle*)&obj->clientNotifyMgrGate);
        }

        /* Close the instance gate if it is opened internally
         * in the instance.
         */
        if ((obj->creGate == TRUE) &&(obj->gate != NULL)) {
            GateMP_close ((GateMP_Handle*)&obj->gate);
        }
        /* Delete create MemMgrs of FrameHeaders anf Frame Buffers.*/
        if (obj->memMgrAllocater_hdrBufs != NULL) {
            SyslinkMemMgr_delete ((SyslinkMemMgr_Handle*)&obj->memMgrAllocater_hdrBufs);
        }

        if (obj->memMgrAllocater_frmBufs != NULL) {
            /* make it based on bufer interface type */
            SyslinkMemMgr_delete ((SyslinkMemMgr_Handle*)&obj->memMgrAllocater_frmBufs);
        }

        break;
    }

    case FrameQBufMgr_ShMem_DYNAMIC_CREATE:
    case FrameQBufMgr_ShMem_DYNAMIC_CREATE_USEDREGION:
    {

        key = GateMP_enter ((GateMP_Handle)obj->gate);

        if (pAttrs != NULL) {
            if (obj->ctrlStructCacheFlag == TRUE) {
                Cache_inv ((Ptr)pAttrs,
                           sizeof (FrameQBufMgr_ShMem_Attrs),
                           Cache_Type_ALL,
                           TRUE);
            }

            for (i = 0; i < pAttrs->numFreeFramePools ; i++) {
                freeFramePoolObj = obj->freeFramePool[i];
                /* Get the frame pointer in virtual format and free the tiler
                 * buffers incase frame buffer is  from tiler
                 */
                for (j = 0; j < freeFramePoolObj->numFrms; j++) {
                    portableAddr = (Ptr)
                    (freeFramePoolObj->frmStartAddrp +
                     (  (   ( (freeFramePoolObj->frmEndAddrp)
                          -  (freeFramePoolObj->frmStartAddrp))
                       /freeFramePoolObj->numFrms)* j) );
                    frame = SyslinkMemMgr_translate (obj->memMgrAllocater_hdrBufs,
                                            (Ptr)portableAddr,
                                             SyslinkMemMgr_AddrType_Portable,
                                             SyslinkMemMgr_AddrType_Virtual);
                    Assert_isTrue((frame != NULL),
                                  FrameQBufMgr_ShMem_A_nullPointer);

              /* Convert frame buffer addresses from portable format to virtual
                * address before returning free frame to the caller.
                */
                frameBufInfo = (FrameQBufMgr_ShMem_FrameBufInfo *)
                                                     &(frame->frameBufInfo[0]);
                for(k = 0; k < frame->numFrameBuffers; k++) {
// DO NOT DELETE
//                         if(obj->bufInterfaceType == FrameQBufMgr_ShMem_INTERFACE_SHAREDMEM){
//                             if (frameBufInfo[k].bufPtr != NULL)
//                             {
//                                 frameBufInfo[k].bufPtr = NULL; // Main purpose
//                             }
//                             /*frame buffer size */
//                             frameBufInfo[k].bufSize = 0;
//                             frmbufOffset +=
//                                         freeFramePoolObj->frmBufSize[k];
//                             /* valid data start offset from buffer start addr */
//                             frameBufInfo[k].startOffset= 0;
//                             /*frame buffer valid size  from the startoffser.
//                              *startoffset+validsize must be less than or equal to bufSize
//                              */
//                             frameBufInfo[k].validSize = 0;
//                         }
                        if(obj->bufInterfaceType == FrameQBufMgr_ShMem_BUFINTERFACE_TILERMEM){
                            if (frameBufInfo[k].bufPtr != NULL)
                            {
                                freeParams.ptr = (Ptr)frameBufInfo[k].bufPtr;
                                SyslinkMemMgr_free(obj->memMgrAllocater_frmBufs, &freeParams);
                                /*TBD for other params*/
                            }
                        }
                    }
                    /*Write back the contents of this frame to memory*/
                    if (obj->frmHdrBufCacheFlag == TRUE) {
                        Cache_wbInv((Ptr)frame,
                                    frame->baseHeaderSize,
                                    TRUE,
                                    NULL);
                    }
                }
                memset((Ptr)freeFramePoolObj,
                        sizeof(FrameQBufMgr_ShMem_FreeFramePoolObj),
                        0);
                if (obj->ctrlStructCacheFlag == TRUE) {
                    Cache_wbInv ((Ptr)freeFramePoolObj,
                                 sizeof (FrameQBufMgr_ShMem_FreeFramePoolObj),
                                 Cache_Type_ALL,
                                 TRUE);
                }
            }

            /* Free the allocated memory for obj->freeFramePool pointers */
            Memory_free (NULL,
                         obj->freeFramePool,
                         sizeof(UInt32)* pAttrs->numFreeFramePools);

            pAttrs->status = !FrameQBufMgr_ShMem_CREATED;
            if (obj->ctrlStructCacheFlag == TRUE) {
                Cache_wbInv ((Ptr)pAttrs,
                             sizeof (FrameQBufMgr_ShMem_Attrs),
                             Cache_Type_ALL,
                             TRUE);
            }

            /* Delete id from the module wide shared info */
            if (obj->instId != (UInt32)-1) {

                instId = obj->instId & 0xFFFF;
                /* Find out the creator procid */
                creProcId = (  (instId >> FRAMEQBUFMGR_CREATORPROCID_BITOFFSET)
                              & FRAMEQBUFMGR_CREATORPROCID_MASK);
                Assert_isTrue ((creProcId < ti_sdo_utils_MultiProc_numProcessors),
                               FrameQBufMgr_A_InvalidCreatorProcId );

                instNo = ( (instId >> FRAMEQBUFMGR_INSTNO_BITOFFSET)
                          & FRAMEQBUFMGR_INSTNO_MASK);
                Assert_isTrue ((instNo < FrameQBufMgr_MAXINSTANCES),
                               FrameQBufMgr_A_InvalidInstNo );
                FrameQBufMgr_freeInstNo(instNo);
            }
        }

        /* Remove entry from NameServer */
        if (   (NULL != FrameQBufMgr_ShMem_module->nameServer)
            && (obj->nsKey != NULL)) {
            NameServer_removeEntry((NameServer_Handle)FrameQBufMgr_ShMem_module->nameServer,
                                   obj->nsKey);
        }

        /*Unregister the notification with the clientNotifyMgr */
        if (obj->isRegistered == TRUE) {
            ClientNotifyMgr_unregisterClient ((ClientNotifyMgr_Handle)obj->clientNotifyMgrHandle,
                                              obj->notifyId);
        }

        if (obj->clientNotifyMgrHandle != NULL) {
            ClientNotifyMgr_delete ((ClientNotifyMgr_Handle *)&(obj->clientNotifyMgrHandle));
        }

        /* Delete GateMP instance of ClientNotifyMgr.*/
        if (   (obj->creCliGate == TRUE)
            && (obj->clientNotifyMgrGate != NULL) ) {
            GateMP_delete ((GateMP_Handle*)&obj->clientNotifyMgrGate);
        }

        GateMP_leave ((GateMP_Handle)obj->gate, key);

        /* Free the allocated memory for the instance control structure
         */
        if (obj->objType == FrameQBufMgr_ShMem_DYNAMIC_CREATE_USEDREGION) {
            Memory_free(SharedRegion_getHeap(obj->regionId),
                        obj->pnameserverEntry,
                        obj->allocSize);
        }

        /* Free up the allocated frame header buffers and
         * frame buffers if they are allocated from shared region.
         */
        if ((obj->hdrBufPtr != NULL) && (obj->hdrBufAllocSize > 0)) {
            Memory_free(SharedRegion_getHeap(obj->hdrBuf_regionId),
                        obj->hdrBufPtr,
                        obj->hdrBufAllocSize);
        }
        if ((obj->frmBufPtr != NULL) && (obj->frmBufAllocSize > 0)) {
            Memory_free(SharedRegion_getHeap(obj->frmBuf_regionId),
                        obj->frmBufPtr,
                        obj->frmBufAllocSize);
        }

        /* Delete instance gate if it is opened internally in the instance.*/
        if ((obj->creGate == TRUE) && (obj->gate != NULL)) {
            GateMP_delete ((GateMP_Handle*)&obj->gate);
            /* Free the allocated memory for the gate instance control
             * structure.
             */
            Memory_free(SharedRegion_getHeap(obj->regionId),
                        obj->instGateSharedAddr,
                        obj->gateMPAllocSize);
        }

        /* Delete created MemMgrs */
        if (obj->memMgrAllocater_hdrBufs != NULL && obj->headerInterfaceType == FrameQBufMgr_ShMem_INTERFACE_SHAREDMEM) {

            SyslinkMemMgr_delete ((SyslinkMemMgr_Handle*)&obj->memMgrAllocater_hdrBufs);
        }

        if (obj->memMgrAllocater_frmBufs != NULL) {
            switch (obj->bufInterfaceType)
            {
                case FrameQBufMgr_ShMem_INTERFACE_SHAREDMEM:
                {
                    SyslinkMemMgr_delete ((SyslinkMemMgr_Handle*)&obj->memMgrAllocater_frmBufs);
                    break;
                }
                case FrameQBufMgr_ShMem_INTERFACE_TILERMEM:
                {
                    SyslinkMemMgr_delete ((SyslinkMemMgr_Handle*)&obj->memMgrAllocater_frmBufs);
                    break;
                }
                default:
                    break;
            }
        }

        /* TODO: Close the heapMemMPinstances/Tiler Mangers opened for MemMgrs
         */
        break;

    }

    case FrameQBufMgr_ShMem_STATIC_CREATE:
    {
        Assert_isTrue (0, FrameQBufMgr_A_invalidParams);
        break;
    }

    case FrameQBufMgr_ShMem_STATIC_OPEN:
    {
        Assert_isTrue (0, FrameQBufMgr_A_invalidParams);
        break;
    }
    default:
            break;
    }
 }

/*
 *  ======== FrameQBufMgr_ShMem_open ========
 *  API to open the created the created instance.
 */
Int32
FrameQBufMgr_ShMem_open(FrameQBufMgr_ShMem_Handle     *handlePtr,
                        ti_syslink_ipc_rtos_shMem_FrameQBufMgr_ShMem_OpenParams *openParams)
{
    Int32                       status = FrameQBufMgr_S_SUCCESS;
    ti_syslink_ipc_rtos_shMem_FrameQBufMgr_ShMem_Params   params;
    Error_Block                 eb;

    Assert_isTrue ((NULL != handlePtr), FrameQBufMgr_ShMem_A_invalidParams);

    /* Reject  if both name and shared Addr is not known */
    if ((handlePtr == NULL) && (openParams == NULL)) {
        status = FrameQBufMgr_E_INVALIDARG;
    }else {
        /* Get the default params */
        if (   strlen((String)openParams->commonOpenParams.name)
             > (FrameQBufMgr_ShMem_MAX_NAMELEN - 1)) {

        }
        else {
            Error_init(&eb);
            ti_syslink_ipc_rtos_shMem_FrameQBufMgr_ShMem_Params_init(&params);
            /* Name is string pointer in CreateParams*/
            params.commonCreateParams.name = (String)
                                             openParams->commonOpenParams.name;
            params.commonCreateParams.openFlag = TRUE;
            params.sharedAddr = openParams->commonOpenParams.sharedAddr;
            /* Opening the instance  */
            params.cpuAccessFlags = openParams->commonOpenParams.cpuAccessFlags;

            *handlePtr = FrameQBufMgr_ShMem_create (&params, &eb);
            if (*handlePtr == NULL) {
                status = FrameQBufMgr_E_FAIL;
            }
        }
    }
    return (status);
}

/*
 *  ======== FrameQBufMgr_ShMem_close ========
 *  Function to close the opened instance .
 */
Int32
FrameQBufMgr_ShMem_close (FrameQBufMgr_ShMem_Handle *handlePtr )
{
    Int32 status = FrameQBufMgr_S_SUCCESS;

    FrameQBufMgr_ShMem_delete (handlePtr);

    return (status);
 }

/*
 *  ======== FrameQBufMgr_ShMem_alloc ========
 *  Function to allocate frame from free frame queue zero .
 */
Int32
FrameQBufMgr_ShMem_alloc (FrameQBufMgr_ShMem_Object *obj,
                          FrameQBufMgr_ShMem_Frame  *framePtr)
{
    Int32                              status           = FrameQBufMgr_S_SUCCESS;
    volatile FrameQBufMgr_ShMem_Attrs  *pAttrs          = NULL;
    Bool                               found            = FALSE;
    FrameQBufMgr_ShMem_FrameBufInfo   *frameBufInfo     = NULL;
    UInt32                             freeFramePoolIndex = 0;
    volatile FrameQBufMgr_ShMem_FreeFramePoolObj     *freeFramePoolObj;
    UInt32                             i;
    UInt32                             key;
    Ptr                                portableAddr;
    FrameQBufMgr_ShMem_Frame           frame;
    Ptr                                virtPtr;
    UInt8                              j;

    Assert_isTrue ((NULL != obj), FrameQBufMgr_ShMem_A_invalidParams);

    Assert_isTrue ((NULL != framePtr), FrameQBufMgr_ShMem_A_invalidParams);

    pAttrs = obj->attrs;

    Assert_isTrue ((NULL != pAttrs), FrameQBufMgr_ShMem_A_invalidParams);

   *framePtr = NULL;
    /* Acquire lock */
    key = GateMP_enter ((GateMP_Handle)obj->gate);
    if (obj->ctrlStructCacheFlag == TRUE) {
        Cache_inv ((Ptr)pAttrs,
                   sizeof (FrameQBufMgr_ShMem_Attrs),
                   Cache_Type_ALL,
                   TRUE);
    }

    /* Alloc allocates frame from the free Frame Pool 0*/
    freeFramePoolIndex = 0;
    freeFramePoolObj = obj->freeFramePool[freeFramePoolIndex];
    if (obj->ctrlStructCacheFlag == TRUE) {
        Cache_inv ((Ptr)freeFramePoolObj,
                   sizeof (FrameQBufMgr_ShMem_FreeFramePoolObj),
                   Cache_Type_ALL,
                   TRUE);
    }

    if (freeFramePoolObj->freeFrms > 0) {
        /* Find out the free frame by checking the reference counts of the
         * frames.
         */
        for (i = 0; i < freeFramePoolObj->numFrms; i++) {
            if(freeFramePoolObj->frmRefCnt[i] == 0) {
                /* free frame found*/
                found = TRUE;
                break;
            }
        }

        if (found == TRUE) {
            freeFramePoolObj->freeFrms -=1u;
            freeFramePoolObj->frmRefCnt[i] += 1u;

            portableAddr = (Ptr)
            (freeFramePoolObj->frmStartAddrp +
             (  (   ( (freeFramePoolObj->frmEndAddrp)
                  -  (freeFramePoolObj->frmStartAddrp))
               /freeFramePoolObj->numFrms)* i) );

                frame = SyslinkMemMgr_translate (obj->memMgrAllocater_hdrBufs,
                                        (Ptr)portableAddr,
                                         SyslinkMemMgr_AddrType_Portable,
                                         SyslinkMemMgr_AddrType_Virtual);
                Assert_isTrue((frame != NULL),
                              FrameQBufMgr_ShMem_A_nullPointer);

               /* Convert frame buffer addresses from portable format to virtual
                * address before returning free frame to the caller.
                */
                frameBufInfo = (FrameQBufMgr_ShMem_FrameBufInfo *)
                                                     &(frame->frameBufInfo[0]);
                for(j = 0; j < frame->numFrameBuffers; j++) {
                    if (obj->bufInterfaceType == FrameQBufMgr_BUFINTERFACE_SHAREDMEM){
                    portableAddr = (Ptr)frameBufInfo[j].bufPtr;
                    virtPtr = SyslinkMemMgr_translate (obj->memMgrAllocater_hdrBufs,
                                        (Ptr)portableAddr,
                                                SyslinkMemMgr_AddrType_Portable,
                                                SyslinkMemMgr_AddrType_Virtual);
                    Assert_isTrue((virtPtr != NULL),
                                   FrameQBufMgr_ShMem_A_nullPointer);
                    frameBufInfo[j].bufPtr = (UInt32)virtPtr;
                }
                    if (obj->bufInterfaceType == FrameQBufMgr_BUFINTERFACE_TILERMEM){
                        /* Currently no translation is required */
                    }
                }
                *framePtr = frame;
        }
        else {
            status     = FrameQBufMgr_E_ALLOC_FRAME;
        }
    }
    else if (freeFramePoolObj->freeDynamicFrames > 0){
        /* TODO:Allocates from the dynamically added frames */
        status      = FrameQBufMgr_E_ALLOC_FRAME;
    }
    else {
        status     = FrameQBufMgr_E_ALLOC_FRAME;
    }

    if (obj->ctrlStructCacheFlag == TRUE) {
       Cache_wbInv ((Ptr)freeFramePoolObj,
                    sizeof (FrameQBufMgr_ShMem_FreeFramePoolObj),
                    Cache_Type_ALL,
                    TRUE);
    }

    if (((status < 0)||(found == FALSE))&& (obj->isRegistered == TRUE)) {
        ClientNotifyMgr_enableNotification ((ClientNotifyMgr_Handle)obj->clientNotifyMgrHandle,
                                            obj->notifyId,
                                            freeFramePoolIndex);
    }

    GateMP_leave((GateMP_Handle)obj->gate, key);

    return (status);
}

/*
 *  ======== FrameQBufMgr_ShMem_allocv ========
 *  Function to allocate multiple frames from the FrameQbufMgr instance.
 */
Int32
FrameQBufMgr_ShMem_allocv (FrameQBufMgr_ShMem_Object *obj,
                           FrameQBufMgr_ShMem_Frame  framePtr[],
                           UInt32                    freeQId[],
                           UInt8                     *numFrames)
{
    Int32                              status           = FrameQBufMgr_S_SUCCESS;
    volatile FrameQBufMgr_ShMem_Attrs  *pAttrs          = NULL;
    Bool                               found            = FALSE;
    FrameQBufMgr_ShMem_FrameBufInfo   *frameBufInfo = NULL;
    UInt32                             lframes      = 0;
    UInt32                             freeFramePoolIndex;
    volatile FrameQBufMgr_ShMem_FreeFramePoolObj     *freeFramePoolObj;
    UInt32                             i;
    UInt32                             key;
    Ptr                                portableAddr;
    FrameQBufMgr_ShMem_Frame           frame;
    Ptr                                virtPtr;
    UInt8                              j;
    Int32                              tmpStatus;
    /* TO find out the duplicate freeQ ids and frames requested in that free Q*/
    UInt32                             pfreeQIds[FrameQBufMgr_MAX_FRAMESINVAPI] = {0};/* TBD: 64 needs to be a macro*/
    UInt32                             pNumFrames[FrameQBufMgr_MAX_FRAMESINVAPI] = {0};/* TBD: 64 needs to be a macro*/
    UInt32                             count = 0;
    UInt32                             pNotifySubIds[FrameQBufMgr_MAX_FRAMESINVAPI];/* TBD: 64 needs to be a macro*/
    UInt32                             numSubEntries = 0;
    Error_Block                        eb;

    Assert_isTrue ((NULL != obj), FrameQBufMgr_ShMem_A_invalidParams);

    Assert_isTrue ((NULL != framePtr), FrameQBufMgr_ShMem_A_invalidParams);

    Assert_isTrue ((NULL != freeQId), FrameQBufMgr_ShMem_A_invalidParams);

    Assert_isTrue ((NULL != numFrames), FrameQBufMgr_ShMem_A_invalidParams);

    Assert_isTrue ((FrameQBufMgr_ShMem_MAX_POOLS >= *numFrames),
                    FrameQBufMgr_ShMem_A_invalidParams);

    Assert_isTrue ((FrameQBufMgr_MAX_FRAMESINVAPI >= *numFrames),
                    FrameQBufMgr_ShMem_A_invalidParams);

    Error_init(&eb);

    /* Find out the duplicate entries in freeQids specified */
    for (i = 0 ; i < *numFrames ; i++) {
        if (pfreeQIds [i] != (Uint32) -1) {
            pfreeQIds [count] = freeQId[i];
            pNumFrames [count] += 1;

            for (j = i + 1 ; j < *numFrames ; j++) {
                if (pfreeQIds [count] ==  freeQId[j]) {
                    pNumFrames [count] += 1;
                    pfreeQIds [j] = (Uint32) -1 ;
                }
            }
            count++ ;
        }
    }

    pAttrs = obj->attrs;

    /* Acquire lock */
    key = GateMP_enter ((GateMP_Handle)obj->gate);

    if (obj->ctrlStructCacheFlag == TRUE) {
        Cache_inv ((Ptr)pAttrs,
                   sizeof (FrameQBufMgr_ShMem_Attrs),
                   Cache_Type_ALL,
                   TRUE);
    }

    for (i = 0; i < count; i++) {
        freeFramePoolIndex = pfreeQIds[i];
        freeFramePoolObj = obj->freeFramePool[freeFramePoolIndex];
        if (obj->ctrlStructCacheFlag == TRUE) {
            Cache_inv ((Ptr)freeFramePoolObj,
                       sizeof(FrameQBufMgr_ShMem_FreeFramePoolObj),
                       Cache_Type_ALL,
                       TRUE);
        }
        if (  ( (freeFramePoolObj->freeFrms)
                +(freeFramePoolObj->freeDynamicFrames ))
              < pNumFrames [i]) {
             status = FrameQBufMgr_E_ALLOC_FRAME;
             /* We are using subNotifyentry 0 for framePool 0 and on */
             pNotifySubIds[numSubEntries] = freeFramePoolIndex;
             numSubEntries +=1;

        }
    }

    if (status >= 0) {
        for (i = 0;  i < *numFrames; i++) {
            /* Allocates frame from the  Frame Pool*/
            freeFramePoolIndex = freeQId[i];
            freeFramePoolObj = obj->freeFramePool[freeFramePoolIndex];
            if (freeFramePoolObj->freeFrms > 0) {
                /* Find out the free frame by checking the reference counts of the
                 * frames
                 */
                for (j = 0;
                     j < freeFramePoolObj->numFrms;
                     j++) {
                    if(   freeFramePoolObj->frmRefCnt[j]
                        == 0) {
                        /* free frame found*/
                        found = TRUE;
                        break;
                    }
                }

                if (found == TRUE) {
                    freeFramePoolObj->freeFrms -=1u;
                    freeFramePoolObj->frmRefCnt[j] += 1u;
                    portableAddr = (Ptr)
                    (freeFramePoolObj->frmStartAddrp +
                     (  (   ( (freeFramePoolObj->frmEndAddrp)
                          -  (freeFramePoolObj->frmStartAddrp))
                       /freeFramePoolObj->numFrms)* j) );

                    frame = SyslinkMemMgr_translate (
                                               obj->memMgrAllocater_hdrBufs,
                                               (Ptr)portableAddr,
                                               SyslinkMemMgr_AddrType_Portable,
                                               SyslinkMemMgr_AddrType_Virtual);
                    /* Convert frame buffer addresses from portable
                     * format to virtual address before returning free
                     * frame to the caller.
                     */
                    frameBufInfo = (FrameQBufMgr_ShMem_FrameBufInfo *)
                                   &(frame->frameBufInfo[0]);
                    for(j = 0; j < frame->numFrameBuffers; j++) {
                        if (obj->bufInterfaceType == FrameQBufMgr_BUFINTERFACE_SHAREDMEM){
                        portableAddr = (Ptr)frameBufInfo[j].bufPtr;
                            virtPtr = SyslinkMemMgr_translate (
                                           obj->memMgrAllocater_hdrBufs,
                                           (Ptr)portableAddr,
                                               SyslinkMemMgr_AddrType_Portable,
                                               SyslinkMemMgr_AddrType_Virtual);
                        Assert_isTrue((virtPtr != NULL),
                                      FrameQBufMgr_ShMem_A_nullPointer);
                        frameBufInfo[j].bufPtr = (UInt32)virtPtr;
                    }
                        if (obj->bufInterfaceType == FrameQBufMgr_BUFINTERFACE_TILERMEM){
                            /* Currently no translation is required */
                        }
                    }
                    framePtr[i] = frame;
                    lframes += 1;
                }
            }
            else if (freeFramePoolObj->freeDynamicFrames > 0){
                /* TODO:Allocates from the dynamically added frames */
                status        = FrameQBufMgr_E_ALLOC_FRAME;
                framePtr[i]  = NULL;
            }
            else {
                status        = FrameQBufMgr_E_ALLOC_FRAME;
                framePtr[i]  = NULL;

            }

           if (obj->ctrlStructCacheFlag == TRUE) {
               Cache_wbInv ((Ptr)freeFramePoolObj,
                            sizeof (FrameQBufMgr_ShMem_FreeFramePoolObj),
                            Cache_Type_ALL,
                            TRUE);
            }

        }
        if ((lframes > 0) && (lframes < *numFrames)) {
            /* Set the status indicating all the frames are not allocated. Only able
             * allocate few frames
             */
            status = FrameQBufMgr_E_ALLOC_ALLFRAMES;
        }
    }

    if (   (   (status == FrameQBufMgr_E_ALLOC_FRAME)
             ||(status == FrameQBufMgr_E_ALLOC_ALLFRAMES))
        && (obj->isRegistered == TRUE)) {
        tmpStatus = ClientNotifyMgr_waitOnMultiSubEntryNotify(
                                                 (ClientNotifyMgr_Handle)obj->clientNotifyMgrHandle,
                                                 obj->notifyId,
                                                 pNotifySubIds,
                                                 numSubEntries);
        if (tmpStatus < 0) {
            status = FrameQBufMgr_E_FAIL;
        }
    }

    if (lframes == 0) {
        for (i = 0; i < *numFrames; i++) {
            framePtr[i]  = NULL;
        }
    }

    /* Update the *numFrames with the  number of frames allocated */
    *numFrames = lframes;

    GateMP_leave((GateMP_Handle)obj->gate, key);

    return (status);
}

/*
 *  ======== FrameQBufMgr_ShMem_free ========
 *  Function to free frame.
 *
 *  This API picks up the freeFrmQueueNo from the frame
 *  Header and directly goes the frame pool to free the frame.
 */
Int32
FrameQBufMgr_ShMem_free (FrameQBufMgr_ShMem_Object *obj,
                         FrameQBufMgr_ShMem_Frame   frame)
{
    Int32                         status        = FrameQBufMgr_S_SUCCESS;
    Bool                          toNotify      = FALSE;
#if defined (SYSLINK_FRAMEQBUFMGR_SUPPORT_DYN_ADD_FRAMES)
    Bool                          freed         = FALSE;
#endif
    UInt32                        index         = 0;
    UInt32                        i             = 0;
    UInt32                        totalFreeFrames ;
    FrameQBufMgr_ShMem_FrameBufInfo              *frameBufInfo  = NULL;
    volatile FrameQBufMgr_ShMem_FreeFramePoolObj *freeFramePoolObj;
    volatile FrameQBufMgr_ShMem_Attrs  *pAttrs  = NULL;
    UInt32                        portablePtr;
    UInt32                        key;
    UInt32                        refIndex;

    Assert_isTrue ((NULL != obj), FrameQBufMgr_ShMem_A_invalidParams);

    Assert_isTrue ((NULL != frame), FrameQBufMgr_ShMem_A_invalidParams);

    Assert_isTrue ((NULL != obj->clientNotifyMgrHandle) ,
                    FrameQBufMgr_A_invalidParams);

    if ((frame == NULL) || (obj == NULL)) {
        status = FrameQBufMgr_E_INVALIDARG;
    }
    else {
        pAttrs = obj->attrs;

        key = GateMP_enter ((GateMP_Handle)obj->gate);

        if (obj->ctrlStructCacheFlag == TRUE) {
            Cache_inv ((Ptr)pAttrs,
                       sizeof (FrameQBufMgr_ShMem_Attrs),
                       Cache_Type_ALL,
                       TRUE);
        }

        index = frame->freeFrmQueueNo;

        if (index >= pAttrs->numFreeFramePools) {
            status = FrameQBufMgr_E_FAIL;
        }
        else {

            /* Convert the frame addr to portable address to check  to which frame
             * pool it belongs.
             */
            portablePtr = (UInt32)SyslinkMemMgr_translate (
                                            obj->memMgrAllocater_hdrBufs,
                                            (Ptr)frame,
                                            SyslinkMemMgr_AddrType_Virtual,
                                            SyslinkMemMgr_AddrType_Portable);
            Assert_isTrue((portablePtr != NULL),
                           FrameQBufMgr_ShMem_A_nullPointer);

            freeFramePoolObj = obj->freeFramePool[index];
            if (obj->ctrlStructCacheFlag == TRUE) {
                Cache_inv ((Ptr)freeFramePoolObj,
                           sizeof (FrameQBufMgr_ShMem_FreeFramePoolObj),
                           Cache_Type_ALL,
                           TRUE);
            }

            /* Check if frame  belongs to the frames created during create time
             */
            if (  (portablePtr >= freeFramePoolObj->frmStartAddrp)
                &&(portablePtr <  freeFramePoolObj->frmEndAddrp)) {
                /* Buffer  needs to freed static frame  list*/

                /* Find out the offset of the buffer in static frames to be able
                 * to reduce the reference count
                 */

                 i = ( (portablePtr - (freeFramePoolObj->frmStartAddrp))/
                      freeFramePoolObj->frmHdrBufSize);
                Assert_isTrue((freeFramePoolObj->frmRefCnt[i] != 0),
                           FrameQBufMgr_ShMem_A_invalidFree);
                freeFramePoolObj->frmRefCnt[i]--;
                if (freeFramePoolObj->frmRefCnt[i] == 0) {
                    /* Free buffer is availabe for subsequnet allocs */
                    toNotify = TRUE;
                    freeFramePoolObj->freeFrms++;
                }
                /* convert the frame buffer pointer in this frame to
                 * portable pointers as this  frame header belongs to free frames
                 * and not  belongs to heade buffers.
                 */
                 frameBufInfo = (FrameQBufMgr_ShMem_FrameBufInfo*)
                                                     &(frame->frameBufInfo[0]);
                 for (i = 0; i < frame->numFrameBuffers; i++) {
                    if (obj->bufInterfaceType == FrameQBufMgr_BUFINTERFACE_SHAREDMEM){
                        frameBufInfo[i].bufPtr = (UInt32) SyslinkMemMgr_translate (
                                                    obj->memMgrAllocater_hdrBufs,
                                                    (Ptr)frameBufInfo[i].bufPtr,
                                                        SyslinkMemMgr_AddrType_Virtual,
                                                        SyslinkMemMgr_AddrType_Portable);
                    Assert_isTrue((frameBufInfo[i].bufPtr != NULL),
                                   FrameQBufMgr_ShMem_A_nullPointer);
                    }
                    if(obj->bufInterfaceType == FrameQBufMgr_BUFINTERFACE_TILERMEM){
                        /* Currently no translation is required and frameBuf
                         * should be freed in delete since it was created in
                         * create
                         */
                    }
                 }
                /* WriteBack frame contents to memory  as next alloc should
                 * see the portable frame pointers of Frame buffers.
                 */
                 FrameQBufMgr_ShMem_writeBackHeaderBuf(obj,frame);

            }
            else if ((portablePtr >= freeFramePoolObj->hdrBufStartAddrp)
                &&(portablePtr <  freeFramePoolObj->hdrBufEndAddrp)) {
                /* Check if it is a duped frame */
                i = ( (portablePtr - (freeFramePoolObj->hdrBufStartAddrp))/
                  freeFramePoolObj->frmHdrBufSize);
                /* Make the frame header is free */
                freeFramePoolObj->hdrBufIsUsed[i] = FALSE;

                refIndex = freeFramePoolObj->hdrBuf_refCntIndex[i];

                Assert_isTrue((freeFramePoolObj->frmRefCnt[refIndex] != 0),
                           FrameQBufMgr_ShMem_A_invalidFree);

                freeFramePoolObj->frmRefCnt[refIndex]--;
                if (freeFramePoolObj->frmRefCnt[refIndex] == 0) {
                    /* Now frame becomes free.So notify */
                    toNotify = TRUE;
                    freeFramePoolObj->freeFrms++;
                }
                freeFramePoolObj->freeFrmHdrBufs++;
            }
            else {
#if defined (SYSLINK_FRAMEQBUFMGR_SUPPORT_DYN_ADD_FRAMES)
                /* Check in dynamic frame list */
                for ( i = 0;
                      (   (i < freeFramePoolObj->numDynamicFrames)
                       && (freed == FALSE));
                      i++) {
                    if (portablePtr ==  freeFramePoolObj->dynFrame[i]) {
                        /* Index of the Buffer to be freed is found. */
                        freeFramePoolObj->dynFrmRefCnt[i] -= 1;
                        if (freeFramePoolObj->dynFrmRefCnt[i] == 0) {
                            toNotify = TRUE;
                            freed    = TRUE;
                            freeFramePoolObj->freeDynamicFrames++;
                        }
                        /*convert the frame buffer pointer in this frame to
                         * portable pointers as this  frame header belongs to
                         * dynamic free frames and not belongs to dynamic header
                         * buffers.
                         */
                         frameBufInfo =
                                      (FrameQBufMgr_ShMem_FrameBufInfo*) &(frame->frameBufInfo[0]);
                         for (i = 0; i < frame->numFrameBuffers; i++) {
                            if (obj->bufInterfaceType == FrameQBufMgr_BUFINTERFACE_SHAREDMEM){
                            frameBufInfo[i].bufPtr = (UInt32)
                                       SyslinkMemMgr_translate (
                                                    obj->memMgrAllocater_hdrBufs,
                                                    (Ptr)frameBufInfo[i].bufPtr,
                                                        SyslinkMemMgr_AddrType_Virtual,
                                                        SyslinkMemMgr_AddrType_Portable);
                            Assert_isTrue((frameBufInfo[i].bufPtr != NULL),
                                           FrameQBufMgr_ShMem_A_nullPointer);
                            }
                            if(obj->bufInterfaceType == FrameQBufMgr_BUFINTERFACE_TILERMEM){
                                /* Currently no translation is required and frameBuf
                                 * should be freed in delete since it was created in
                                 * create*/
                            }
                         }

                       break;
                    }
                }
                if (freed == FALSE) {
                    /*Check if buffer is duped with dynamic hdrbufs */
                    /* TODO */
                }
#endif /*#if defined (SYSLINK_FRAMEQBUFMGR_SUPPORT_DYN_ADD_FRAMES)*/
            }
        }

        totalFreeFrames =  (  freeFramePoolObj->freeFrms
                            + freeFramePoolObj->freeDynamicFrames);
        if (obj->ctrlStructCacheFlag == TRUE) {
            Cache_wbInv ((Ptr)freeFramePoolObj,
                         sizeof (FrameQBufMgr_ShMem_FreeFramePoolObj),
                         Cache_Type_ALL,
                         TRUE);
        }

        GateMP_leave ((GateMP_Handle)obj->gate, key);
    }

    if (toNotify == TRUE) {
        ClientNotifyMgr_sendNotificationMulti ( (ClientNotifyMgr_Handle)obj->clientNotifyMgrHandle,
                                                obj->notifyId,
                                                index,
                                                totalFreeFrames,
                                                (UInt32)-1 );

    }
    return (status);
}

/*
 * ======== FrameQBufMgr_ShMem_freev ========
 * Function to free frames more than one.
 *
 * Function to free multiple frames  to the  FrameQBufMgr.Notification to other
 * clients will be generated once all the frames are freed.
 */
Int32
FrameQBufMgr_ShMem_freev (FrameQBufMgr_ShMem_Object *obj,
                          FrameQBufMgr_ShMem_Frame   framePtr[],
                          UInt32                     numFrames)
{
    Int32                         status        = FrameQBufMgr_S_SUCCESS;
    Bool                          toNotify      = FALSE;
    Bool                          freed         = FALSE;
    UInt32                        index         = 0;
    UInt32                        i             = 0;
    UInt32                            totalFreeFrames = 0;
    FrameQBufMgr_ShMem_FrameBufInfo  *frameBufInfo  = NULL;
    volatile FrameQBufMgr_ShMem_FreeFramePoolObj     *freeFramePoolObj;
    volatile FrameQBufMgr_ShMem_Attrs  *pAttrs          = NULL;
    UInt32                        portablePtr;
    UInt32                        key;
    UInt32                        refIndex;
    UInt32                        frmCnt;

    Assert_isTrue ((NULL != obj), FrameQBufMgr_ShMem_A_invalidParams);

    Assert_isTrue ((NULL != framePtr), FrameQBufMgr_ShMem_A_invalidParams);

    Assert_isTrue ((0 != numFrames), FrameQBufMgr_ShMem_A_invalidParams);

    pAttrs = obj->attrs;
    /* Acquire lock */
    key = GateMP_enter ((GateMP_Handle)obj->gate);

    if (obj->ctrlStructCacheFlag == TRUE) {
        Cache_inv ((Ptr)pAttrs,
                   sizeof (FrameQBufMgr_ShMem_ControlStruct),
                   Cache_Type_ALL,
                   TRUE);
    }

    /* Free frames by looking at to which pool no it belongs. */
    for (frmCnt = 0; frmCnt < numFrames; frmCnt++) {
        index = framePtr[frmCnt]->freeFrmQueueNo;

        if (index >= pAttrs->numFreeFramePools) {
            status = FrameQBufMgr_E_FAIL;
        }
        else {
            freeFramePoolObj = obj->freeFramePool[index];
            if (obj->ctrlStructCacheFlag == TRUE) {
                Cache_inv ((Ptr)freeFramePoolObj,
                           sizeof (FrameQBufMgr_ShMem_FreeFramePoolObj),
                           Cache_Type_ALL,
                           TRUE);
            }
            /* Convert frame addr to portable address to check  reference
             * counts etc.
             */
            portablePtr = (UInt32)SyslinkMemMgr_translate (
                                                   obj->memMgrAllocater_hdrBufs,
                                                   (Ptr)framePtr[frmCnt],
                                                   SyslinkMemMgr_AddrType_Virtual,
                                                   SyslinkMemMgr_AddrType_Portable);
            Assert_isTrue((portablePtr != NULL),
                           FrameQBufMgr_ShMem_A_nullPointer);

            /* Check if frame  belongs to the frames created during create time
             */
            if (  (portablePtr >= freeFramePoolObj->frmStartAddrp)
                &&(portablePtr <  freeFramePoolObj->frmEndAddrp)) {
                /* Buffer  needs to freed static frame  list*/

                /* Find out the offset of the buffer in static frames to be able
                 * to reduce the reference count
                 */

                i = ( (portablePtr - (freeFramePoolObj->frmStartAddrp))/
                      freeFramePoolObj->frmHdrBufSize);
                Assert_isTrue((freeFramePoolObj->frmRefCnt[i] != 0),
                           FrameQBufMgr_ShMem_A_invalidFree);

                freeFramePoolObj->frmRefCnt[i]--;
                if (freeFramePoolObj->frmRefCnt[i] == 0) {
                    /* Free buffer is availabe for subsequnet allocs */
                    toNotify = TRUE;
                    freeFramePoolObj->freeFrms++;
                }

                /* Convert the frame buffer pointer in this frame to
                 * portable pointers as this  frame header belongs to free frames
                 * and not  belongs to heade buffers.
                 */
                 frameBufInfo = (FrameQBufMgr_ShMem_FrameBufInfo*)
                                           &(framePtr[frmCnt]->frameBufInfo[0]);
                 for (i = 0; i < framePtr[frmCnt]->numFrameBuffers; i++) {
                    if(obj->bufInterfaceType == FrameQBufMgr_BUFINTERFACE_SHAREDMEM){
                        frameBufInfo[i].bufPtr = (UInt32) SyslinkMemMgr_translate (
                                                   obj->memMgrAllocater_hdrBufs,
                                                  (Ptr)frameBufInfo[i].bufPtr,
                                                       SyslinkMemMgr_AddrType_Virtual,
                                                       SyslinkMemMgr_AddrType_Portable);
                    Assert_isTrue((frameBufInfo[i].bufPtr != NULL),
                                   FrameQBufMgr_ShMem_A_nullPointer);
                    }
                    if(obj->bufInterfaceType == FrameQBufMgr_BUFINTERFACE_TILERMEM){
                        /* Currently no translation is required and frameBuf
                         * should be freed in delete since it was created in
                         * create*/
                    }
                 }
                /* WriteBack frame contents to memory  as next alloc should
                 * see the portable frame pointers of Frame buffers.
                 */
                 FrameQBufMgr_ShMem_writeBackHeaderBuf(obj, framePtr[frmCnt]);
            }
            else if ((portablePtr >= freeFramePoolObj->hdrBufStartAddrp)
                &&(portablePtr <  freeFramePoolObj->hdrBufEndAddrp)) {
                /* Check if it is a duped frame */
                i = ( (portablePtr - (freeFramePoolObj->hdrBufStartAddrp))/
                  freeFramePoolObj->frmHdrBufSize);
                /* Make the frame header is free */
                freeFramePoolObj->hdrBufIsUsed[i] = FALSE;

                refIndex = freeFramePoolObj->hdrBuf_refCntIndex[i];
                Assert_isTrue((freeFramePoolObj->frmRefCnt[refIndex] != 0),
                           FrameQBufMgr_ShMem_A_invalidFree);
                freeFramePoolObj->frmRefCnt[refIndex]--;
                if (freeFramePoolObj->frmRefCnt[refIndex] == 0) {
                    /* Now frame becomes free.So notify */
                    toNotify = TRUE;
                    freeFramePoolObj->freeFrms++;
                }
                freeFramePoolObj->freeFrmHdrBufs++;
            }
            else {
#if defined (SYSLINK_FRAMEQBUFMGR_SUPPORT_DYN_ADD_FRAMES)
                /* Check in dynamic frame list */
                for ( i = 0;
                      (   (i < freeFramePoolObj->numDynamicFrames)
                       && (freed == FALSE));
                      i++) {
                    if (portablePtr ==  freeFramePoolObj->dynFrame[i]) {
                        /* Index of the Buffer to be freed is found. */
                        freeFramePoolObj->dynFrmRefCnt[i] -= 1;
                        if (freeFramePoolObj->dynFrmRefCnt[i] == 0) {
                            toNotify = TRUE;
                            freed    = TRUE;
                            freeFramePoolObj->freeDynamicFrames++;
                        }
                        /*convert the frame buffer pointer in this frame to
                         * portable pointers as this  frame header belongs to
                         * dynamic free frames and not belongs to dynamic header
                         * buffers.
                         */
                         frameBufInfo = (FrameQBufMgr_ShMem_FrameBufInfo*)
                                           &(framePtr[frmCnt]->frameBufInfo[0]);
                         for (i = 0;
                              i < framePtr[frmCnt]->numFrameBuffers;
                              i++) {
                            if(obj->bufInterfaceType == FrameQBufMgr_BUFINTERFACE_SHAREDMEM){
                            frameBufInfo[i].bufPtr = (UInt32)
                                       SyslinkMemMgr_translate (
                                                    obj->memMgrAllocater_hdrBufs,
                                                    (Ptr)frameBufInfo[i].bufPtr,
                                                        SyslinkMemMgr_AddrType_Virtual,
                                                        SyslinkMemMgr_AddrType_Portable);
                            Assert_isTrue((frameBufInfo[i].bufPtr != NULL),
                                           FrameQBufMgr_ShMem_A_nullPointer);
                         }
                            if(obj->bufInterfaceType == FrameQBufMgr_BUFINTERFACE_TILERMEM){
                                /* Currently no translation is required and frameBuf
                                 * should be freed in delete since it was created in
                                 * create*/
                            }
                         }

                       break;
                    }
                }
#endif /*#if defined (SYSLINK_FRAMEQBUFMGR_SUPPORT_DYN_ADD_FRAMES)*/
                if (freed == FALSE) {
                    /*Check if buffer is duped with dynamic hdrbufs */
                    /* TODO */
                }

            }
        }

        totalFreeFrames =  (   freeFramePoolObj->freeFrms
                             + freeFramePoolObj->freeDynamicFrames);
        if (obj->ctrlStructCacheFlag == TRUE) {
           Cache_wbInv ((Ptr)freeFramePoolObj,
                        sizeof (FrameQBufMgr_ShMem_FreeFramePoolObj),
                        Cache_Type_ALL,
                        TRUE);
        }

        if (toNotify == TRUE) {
            /* Release lock */
           GateMP_leave ((GateMP_Handle)obj->gate, key);

            /* Send notification to all the other clients of this instance
             * on this framePool identified by the index.
             */
            ClientNotifyMgr_sendNotificationMulti ((ClientNotifyMgr_Handle)obj->clientNotifyMgrHandle,
                                                   obj->notifyId,
                                                   index,
                                                   totalFreeFrames,
                                                   (UInt32)-1 );
            key = GateMP_enter ((GateMP_Handle)obj->gate);

        }
        toNotify = FALSE;
    }

    GateMP_leave ((GateMP_Handle)obj->gate, key);

    return (status);
}

/*
 *  ======== FrameQBufMgr_ShMem_add ========
 * Function to add  free frames dynamically  which are allocated out side of
 * FrmaeQbufMgr through the  same memory manager.
 */
Ptr
FrameQBufMgr_ShMem_add (FrameQBufMgr_ShMem_Object *obj,
                        UInt8                      freeQId)
{
    Ptr    ptr    = NULL;

    Assert_isTrue ((NULL != obj), FrameQBufMgr_ShMem_A_invalidParams);

    Assert_isTrue ((FrameQBufMgr_MAX_POOLS > freeQId),
                   FrameQBufMgr_ShMem_A_invalidParams);

    /* Currently not implemented */
    return (ptr);
}

/*
 *  ======== FrameQBufMgr_ShMem_remove ========
 * Function to remove frame from the FrameQbufMgr.Only allowed for dynamically
 * added frames.
 */
Int32
FrameQBufMgr_ShMem_remove (FrameQBufMgr_ShMem_Object *obj,
                           UInt8                      freeQId,
                           FrameQBufMgr_ShMem_Frame   framePtr)
{
    Int32  status = FrameQBufMgr_S_SUCCESS;

    Assert_isTrue ((NULL != obj), FrameQBufMgr_ShMem_A_invalidParams);

    Assert_isTrue ((FrameQBufMgr_MAX_POOLS > freeQId),
                    FrameQBufMgr_ShMem_A_invalidParams);

    Assert_isTrue ((NULL != framePtr), FrameQBufMgr_ShMem_A_invalidParams);

    /* Currently not implemented */
    status = FrameQBufMgr_E_ACCESSDENIED;

    return (status);
}

/*
 *  ======== FrameQBufMgr_ShMem_registerNotifier ========
 * Function to Register notification with the instance.
 *
 * It internally registers notification with the individual free queues(free frame pools).
 * If alloc on free queue 0 failed,Notification will be generated if free frames
 * in that queue becomes more than watermark(all the free frames in that queue).
 */
Int32
FrameQBufMgr_ShMem_registerNotifier (
                               FrameQBufMgr_ShMem_Object       *obj,
                               FrameQBufMgr_ShMem_NotifyParams *notifyParams)
{
    Int32                                  status = FrameQBufMgr_S_SUCCESS;
    ClientNotifyMgr_registerNoitifyParams  regParams;
    UInt32                      lNotifyId;

    Assert_isTrue ((NULL != obj), FrameQBufMgr_ShMem_A_invalidParams);

    Assert_isTrue ((NULL != notifyParams), FrameQBufMgr_ShMem_A_invalidParams);

    /* Register call back function with the clientNotifyMgr.Pass the
     * cbClientHandle which is passed by the caller of this API to
     * clientNotifyMgr so that it will be passed as first arg to the call back
     * function during notification.
     */
    regParams.clientHandle   = notifyParams->cbClientHandle;
    regParams.procId         = MultiProc_self();
    regParams.notifyType     = notifyParams->notifyType;
    regParams.fxnPtr         = (ClientNotifyMgr_FnCbck)(notifyParams->cbFxnPtr);
    regParams.cbContext      = notifyParams->cbContext;
    regParams.watermarkCond1 = notifyParams->watermark;
    regParams.watermarkCond2 = (UInt32)-1;

    /*Call ClientNotifyMgr_registerClient API */
    status = ClientNotifyMgr_registerClient((ClientNotifyMgr_Handle)obj->clientNotifyMgrHandle,
                                            &regParams,
                                            &lNotifyId);
    if (status >= 0) {
        obj->isRegistered = TRUE;
        obj->notifyId = lNotifyId;
    }

    return (status);

}

/*
 *  ======== FrameQBufMgr_ShMem_unregisterNotifier ========
 * Function to unregister the notification call back function.
 */
Int32
FrameQBufMgr_ShMem_unregisterNotifier (FrameQBufMgr_ShMem_Object *obj)
{
    Int32  status = FrameQBufMgr_S_SUCCESS;

    Assert_isTrue ((NULL != obj), FrameQBufMgr_ShMem_A_invalidParams);

    /*Call ClientNotifyMgr_unregisterClient API */
    status = ClientNotifyMgr_unregisterClient ((ClientNotifyMgr_Handle)obj->clientNotifyMgrHandle,
                                               obj->notifyId);
    if (status >= 0) {
        obj->isRegistered = FALSE;
    }

    return (status);
}

/*
 *  ======== FrameQBufMgr_ShMem_dup ========
 * API to duplicate the given frame i.e it increments the corresponding
 * frames 's reference count.It internally allocates  headers for the duped
 * frames.It fails if it is not able to dup all the frames.
 */
Int32
FrameQBufMgr_ShMem_dup (FrameQBufMgr_ShMem_Object *obj,
                        FrameQBufMgr_ShMem_Frame   frame,
                        FrameQBufMgr_ShMem_Frame   dupedFramePtr[],
                        UInt32                     numDupedFrames)
{
    Int32   status = FrameQBufMgr_S_SUCCESS;
    Bool    bDuped = FALSE;
    UInt32  ldupedFrames = 0;
    volatile FrameQBufMgr_ShMem_Attrs  *pAttrs          = NULL;
    volatile FrameQBufMgr_ShMem_FreeFramePoolObj *freeFramePoolObj;
    FrameQBufMgr_ShMem_Frame lframe;
    UInt32  portablePtr;
    UInt8   index;
    UInt32  lFrameIndex;
    UInt8   indexType;
    UInt32  refIndex;
    UInt32  i;
    UInt32  k;
    Int32   key;

    Assert_isTrue ((NULL != obj), FrameQBufMgr_ShMem_A_invalidParams);

    Assert_isTrue ((NULL != frame), FrameQBufMgr_ShMem_A_invalidParams);

    Assert_isTrue ((NULL != dupedFramePtr), FrameQBufMgr_ShMem_A_invalidParams);

    Assert_isTrue ((0u != numDupedFrames), FrameQBufMgr_ShMem_A_invalidParams);

    index = frame->freeFrmQueueNo;
    /* Reset numDupedFramePtr*/
    for (i = 0; i < numDupedFrames; i++) {
        dupedFramePtr[i] = NULL;
    }

    pAttrs = obj->attrs;

    /* Acquire lock */
    key = GateMP_enter ((GateMP_Handle)obj->gate);

    if (obj->ctrlStructCacheFlag == TRUE) {
        Cache_inv ((Ptr)pAttrs,
                   sizeof (FrameQBufMgr_ShMem_ControlStruct),
                   Cache_Type_ALL,
                   TRUE);
    }

    /* Check if the frame is static frame or duped frame created from the static
     * headers or dynamic frame or frame with duped dynamic header.
     */
    if (index >= pAttrs->numFreeFramePools) {
        status = FrameQBufMgr_E_FAIL;
    }
    else {
        freeFramePoolObj = obj->freeFramePool[index];
        if (obj->ctrlStructCacheFlag == TRUE) {
            Cache_inv ((Ptr)freeFramePoolObj,
                       sizeof (FrameQBufMgr_ShMem_FreeFramePoolObj),
                       Cache_Type_ALL,
                       TRUE);
        }
        /* Check if this queue has sufficient number of free header bufs to dup
         * the given frame.
         */
         if (    numDupedFrames
              > (  freeFramePoolObj->freeFrmHdrBufs
                 + freeFramePoolObj->freeDynamicHdrBufs)) {
             /* Can not satisfy the requirement with the available free header
              * buffers.
              */
             status = FrameQBufMgr_E_INSUFFICIENT_HDRBUFS;
         }
         else {
            /* Convert frame addr to portable address to check  reference
             * counts etc.
             */
            portablePtr = (UInt32)SyslinkMemMgr_translate (
                                            obj->memMgrAllocater_hdrBufs,
                                            (Ptr)frame,
                                            SyslinkMemMgr_AddrType_Virtual,
                                            SyslinkMemMgr_AddrType_Portable);
            Assert_isTrue((portablePtr != NULL),
                           FrameQBufMgr_ShMem_A_nullPointer);

            /* Check if frame  belongs to the frames created during create time
             */
            if (  (portablePtr >= freeFramePoolObj->frmStartAddrp)
                &&(portablePtr <  freeFramePoolObj->frmEndAddrp)) {
                /* Given frame belons to static frame.*/

                /* Find out the offset of the buffer in static frames to be able
                 * to update the reference count of the given frame.
                 */
                 i = ( (portablePtr - (freeFramePoolObj->frmStartAddrp))/
                      freeFramePoolObj->frmHdrBufSize);
                 /* Increment reference count by numDupedFrames */
                 freeFramePoolObj->frmRefCnt[i] += numDupedFrames;
                 lFrameIndex = i;
                /* Indicates original frame that is pointed to is
                 * static frame.
                 */
                indexType = FrameQBufMgr_ShMem_STATIC_FRAME;

            }
            else if (  (portablePtr >= freeFramePoolObj->hdrBufStartAddrp)
                     &&(portablePtr <  freeFramePoolObj->hdrBufEndAddrp))
                {
                /* Check if the given frame is a duped frame  with static
                 * header .
                 */
                i = ( (portablePtr - (freeFramePoolObj->hdrBufStartAddrp))/
                  freeFramePoolObj->frmHdrBufSize);

                /* Find out the original frame to which the given duped frame
                 * is referring to.
                 */
                refIndex = freeFramePoolObj->hdrBuf_refCntIndex[i];
                lFrameIndex = refIndex;
                /* Check if this Header is pointing to dynamic frame or
                 * static frame.
                 */
                if (freeFramePoolObj->hdrBuf_refCntIndexType[i] ==
                                              FrameQBufMgr_ShMem_STATIC_FRAME) {
                    freeFramePoolObj->frmRefCnt[refIndex] += numDupedFrames;
                    /* Indicates original frame that is pointed to is
                     * static frame.
                     */
                    indexType = FrameQBufMgr_ShMem_STATIC_FRAME;
                }
                else {
                    /* This header buffers is actually duped frame for frame
                     * whcih belongs dynamic frames of this queue
                     */
                    freeFramePoolObj->dynFrmRefCnt[refIndex] += numDupedFrames;
                    /* Indicates original frame that is pointed to is
                     * dynamic frame.
                     */
                    indexType = FrameQBufMgr_ShMem_DYNAMIC_FRAME;
                }

            }
            else {
                /* Check in dynamic frame list */
                for ( i = 0;
                      (i < freeFramePoolObj->numDynamicFrames);
                      i++) {
                    if (portablePtr ==  freeFramePoolObj->dynFrame[i]) {
                        /* Index of the Buffer to be freed is found. */
                        freeFramePoolObj->dynFrmRefCnt[i] += numDupedFrames;
                        bDuped = TRUE;
                        lFrameIndex = i;
                        /* Indicates original frame that is pointed to is
                         * dynamic frame.
                         */
                        indexType = FrameQBufMgr_ShMem_DYNAMIC_FRAME;

                    }
                }

                if (bDuped == FALSE) {
                    /* Check if given frame header is from dynamic frame
                     * headers
                     */
                    for ( i = 0;
                          (i < freeFramePoolObj->numDynamicHdrBufs);
                          i++) {
                        if (    portablePtr
                             == freeFramePoolObj->dynHdrBufs[i]) {
                            /* Find out the original frame to which the given duped frame
                             * is referring to.
                             */
                            refIndex =
                               freeFramePoolObj->dynHdrBuf_refCntIndex[i];
                            lFrameIndex = refIndex;
                            /* Check if this Header is pointing to dynamic frame or
                             * static frame and increment ref count of static
                             * frame or dynamic frame.
                             */
                            if (   (freeFramePoolObj->
                                    dynHdrBuf_refCntIndexType[i])
                                ==  (FrameQBufMgr_ShMem_STATIC_FRAME)) {
                                freeFramePoolObj->frmRefCnt[refIndex] +=
                                                                 numDupedFrames;
                                /* Indicates original frame that is pointed to
                                 * is static frame.
                                 */
                                indexType = FrameQBufMgr_ShMem_STATIC_FRAME;

                            }
                            else {
                                /* This header buffers is actually duped frame for frame
                                 * whcih belongs dynamic frames of this queue
                                 */
                                freeFramePoolObj->dynFrmRefCnt[refIndex]
                                                              += numDupedFrames;
                                /* Indicates original frame that is pointed to
                                 * is dynamic frame.
                                 */
                                indexType = FrameQBufMgr_ShMem_DYNAMIC_FRAME;

                            }

                        }
                    }
                } /*if (bDuped == FALSE) {*/
            }
            /* Now Allocate  duped headers  from the quque and copy the contents
             * of the orginal frame to duped frames  and  return
             */
            if (status >= 0) {
                for (k = 0;
                     (  (k < freeFramePoolObj->numFrmHdrBufs)
                      &&(ldupedFrames < numDupedFrames)
                      &&(freeFramePoolObj->freeFrmHdrBufs > 0));
                     k++) {
                    if (freeFramePoolObj->hdrBufIsUsed[k] != TRUE){
                        /* Found free header */
                        freeFramePoolObj->hdrBufIsUsed[k] = TRUE;

                        freeFramePoolObj->hdrBuf_refCntIndex[k] =
                                                                   lFrameIndex;
                        /* Set if it is pointing to static frame or
                         * dynamic frame.
                         */
                        freeFramePoolObj->hdrBuf_refCntIndexType[k] =
                                                                     indexType;
                        freeFramePoolObj->freeFrmHdrBufs--;
                        portablePtr = (UInt32)
                          (   freeFramePoolObj->hdrBufStartAddrp
                            + (  (   (  (freeFramePoolObj->hdrBufEndAddrp)
                                       -(freeFramePoolObj->hdrBufStartAddrp))
                                   / (freeFramePoolObj->numFrmHdrBufs))
                                * k) );

                        lframe = SyslinkMemMgr_translate (obj->memMgrAllocater_hdrBufs,
                                                   (Ptr)portablePtr,
                                                   SyslinkMemMgr_AddrType_Portable,
                                                   SyslinkMemMgr_AddrType_Virtual);
                        /* Copy  given frame header contens to this frame*/
                        memcpy(lframe, frame,  frame->headerSize);
                        dupedFramePtr[ldupedFrames] = lframe;
                        ldupedFrames++;
                    }

                }/*for (k = 0;  */

                if (ldupedFrames < numDupedFrames) {
                    /* If still  need to allocate */
                    for (k = 0;
                        (  (k < freeFramePoolObj->numDynamicHdrBufs)
                         &&(ldupedFrames < numDupedFrames)
                         &&(freeFramePoolObj->freeDynamicHdrBufs > 0));
                        k++) {
                        if (freeFramePoolObj->dynHdrBufIsUsed[k] != TRUE){
                            /* Found free header */
                            freeFramePoolObj->dynHdrBufIsUsed[k] = TRUE;

                            freeFramePoolObj->dynHdrBuf_refCntIndex[k] =
                                                                    lFrameIndex;
                            /* Set if it is pointing to static frame or
                             * dynamic frame.
                             */
                            freeFramePoolObj->dynHdrBuf_refCntIndexType[k]
                                                                   = indexType;
                            freeFramePoolObj->freeDynamicHdrBufs--;
                            portablePtr = (UInt32)
                                          freeFramePoolObj->dynHdrBufs[k];

                            lframe = SyslinkMemMgr_translate (
                                                   obj->memMgrAllocater_hdrBufs,
                                                   (Ptr)portablePtr,
                                                   SyslinkMemMgr_AddrType_Portable,
                                                   SyslinkMemMgr_AddrType_Virtual);
                            /* Copy  given frame header contens to this frame*/
                            memcpy(lframe, frame,  frame->headerSize);
                            dupedFramePtr[ldupedFrames] = lframe;
                            ldupedFrames++;
                        }

                    }/*for (k = 0;  */

                }/*if (ldupedFrames < numDupedFrames) */
            }
        }

        /* Writeback contents of the frame pool */
        if (obj->ctrlStructCacheFlag == TRUE) {
            Cache_wbInv ((Ptr)freeFramePoolObj,
                         sizeof (FrameQBufMgr_ShMem_FreeFramePoolObj),
                         Cache_Type_ALL,
                         TRUE);
        }

    }

    /* Release lock */
    GateMP_leave ((GateMP_Handle)obj->gate, key);

    return (status);
}

/*
 *  ======== FrameQBufMgr_ShMem_dupv ========
 * API to duplicate the given frames i.e it increments the corresponding
 * frames 's reference count.It internally allocates  headers for the duped
 * frames.It fails if it is not able to dup all the frames.
 */
Int32
FrameQBufMgr_ShMem_dupv (FrameQBufMgr_ShMem_Object  *obj,
                         FrameQBufMgr_ShMem_Frame    framePtr[],
                         FrameQBufMgr_ShMem_Frame  **dupedFramePtr,
                         UInt32                      numDupedFrames,
                         UInt32                      numFrames)
{
    Int32       status = FrameQBufMgr_S_SUCCESS;
    Bool        bDuped = FALSE;
    UInt32      ldupedFrames = 0;
    UInt8       indexType    = 0xFF;
    volatile FrameQBufMgr_ShMem_Attrs  *pAttrs          = NULL;
    volatile FrameQBufMgr_ShMem_FreeFramePoolObj *freeFramePoolObj;
    FrameQBufMgr_ShMem_Frame lframe;
    UInt32      portablePtr;
    UInt8       index;
    UInt32      lFrameIndex;
    UInt32      lFrmCnt;
    UInt32      refIndex;
    UInt32      i;
    UInt32      k;
    Int32       key;

    Assert_isTrue ((NULL != obj), FrameQBufMgr_ShMem_A_invalidParams);

    Assert_isTrue ((NULL != framePtr), FrameQBufMgr_ShMem_A_invalidParams);

    Assert_isTrue ((NULL != dupedFramePtr), FrameQBufMgr_ShMem_A_invalidParams);

    Assert_isTrue ((0 != numDupedFrames), FrameQBufMgr_ShMem_A_invalidParams);

    Assert_isTrue ((0 != numFrames), FrameQBufMgr_ShMem_A_invalidParams);

    pAttrs = obj->attrs;
    /* Acquire lock */
    key = GateMP_enter ((GateMP_Handle)obj->gate);

    if (obj->ctrlStructCacheFlag == TRUE) {
        Cache_inv ((Ptr)pAttrs,
                   sizeof (FrameQBufMgr_ShMem_Attrs),
                   Cache_Type_ALL,
                   TRUE);
    }

    /* Return error if header buffers  available are not sufficient*/
    for (lFrmCnt = 0; (lFrmCnt < numFrames) && (status >= 0); lFrmCnt++) {
        index = framePtr[lFrmCnt]->freeFrmQueueNo;
        /* Check if the frame is static frame or duped frame created from the static
         * headers or dynamic frame or frame with duped dynamic header.
         */
        if (index >= pAttrs->numFreeFramePools) {
            status = FrameQBufMgr_E_FAIL;
        }
        else {
            freeFramePoolObj = obj->freeFramePool[index];
            if (obj->ctrlStructCacheFlag == TRUE) {
                Cache_inv ((Ptr)freeFramePoolObj,
                           sizeof (FrameQBufMgr_ShMem_FreeFramePoolObj),
                           Cache_Type_ALL,
                           TRUE);
            }
            /* Check if this queue has sufficient number of free header bufs to dup
             * the given frame.
             */
            if (    numDupedFrames
                  > (  freeFramePoolObj->freeFrmHdrBufs
                     + freeFramePoolObj->freeDynamicHdrBufs)) {
                 /* Can not satisfy the requirement with the available free header
                  * buffers.
                  */
                status = FrameQBufMgr_E_INSUFFICIENT_HDRBUFS;
            }
        }
    }

    if (status >= 0) {
        for (lFrmCnt = 0; lFrmCnt < numFrames; lFrmCnt++) {
            index = framePtr[lFrmCnt]->freeFrmQueueNo;
            ldupedFrames = 0;

            /* Convert frame addr to portable address to check  reference
             * counts etc.
             */
            portablePtr = (UInt32)SyslinkMemMgr_translate (
                                                   obj->memMgrAllocater_hdrBufs,
                                                   (Ptr)framePtr[lFrmCnt],
                                                    SyslinkMemMgr_AddrType_Virtual,
                                                    SyslinkMemMgr_AddrType_Portable);
            Assert_isTrue((portablePtr != NULL),
                           FrameQBufMgr_ShMem_A_nullPointer);

            /* Check if frame  belongs to the frames created during create time
             */
            if (  (portablePtr >= freeFramePoolObj->frmStartAddrp)
                &&(portablePtr <  freeFramePoolObj->frmEndAddrp)) {
                /* Given frame belons to static frame.*/

                /* Find out the offset of the buffer in static frames to be able
                 * to update the reference count of the given frame.
                 */
                 i = ( (portablePtr - (freeFramePoolObj->frmStartAddrp))/
                      freeFramePoolObj->frmHdrBufSize);
                 /* Increment reference count by numDupedFrames */
                 freeFramePoolObj->frmRefCnt[i] += numDupedFrames;
                 lFrameIndex = i;
                /* Indicates original frame that is pointed to is
                 * static frame.
                 */
                indexType = FrameQBufMgr_ShMem_STATIC_FRAME;

            }
            else if (  (portablePtr >= freeFramePoolObj->hdrBufStartAddrp)
                     &&(portablePtr <  freeFramePoolObj->hdrBufEndAddrp))
                {
                /* Check if the given frame is a duped frame  with static
                 * header .
                 */
                i = ( (portablePtr - (freeFramePoolObj->hdrBufStartAddrp))/
                  freeFramePoolObj->frmHdrBufSize);

                /* Find out the original frame to which the given duped frame
                 * is referring to.
                 */
                refIndex = freeFramePoolObj->hdrBuf_refCntIndex[i];
                lFrameIndex = refIndex;
                /* Check if this Header is pointing to dynamic frame or
                 * static frame.
                 */
                if (freeFramePoolObj->hdrBuf_refCntIndexType[i] ==
                                              FrameQBufMgr_ShMem_STATIC_FRAME) {
                    freeFramePoolObj->frmRefCnt[refIndex] +=
                                                                numDupedFrames;
                    /* Indicates that original frame is
                     * static frame.
                     */
                    indexType = FrameQBufMgr_ShMem_STATIC_FRAME;
                }
                else {
                    /* This header buffers is actually duped frame for frame
                     * whcih belongs dynamic frames of this queue
                     */
                    freeFramePoolObj->dynFrmRefCnt[refIndex] +=
                                                               numDupedFrames;
                    /* Indicates that original frame  is
                     * dynamic frame.
                     */
                    indexType = FrameQBufMgr_ShMem_DYNAMIC_FRAME;
                }

            }
            else {
                /* Check in dynamic frame list */
                for ( i = 0;
                      (i < freeFramePoolObj->numDynamicFrames);
                      i++) {
                    if (portablePtr ==  freeFramePoolObj->dynFrame[i]) {
                        /* Index of the Buffer to be freed is found. */
                        freeFramePoolObj->dynFrmRefCnt[i] += numDupedFrames;
                        bDuped = TRUE;
                        lFrameIndex = i;
                        indexType = FrameQBufMgr_ShMem_DYNAMIC_FRAME;

                    }
                }

                if (bDuped == FALSE) {
                    /* Check if given frame header is from dynamic frame
                     * headers
                     */
                    for ( i = 0;
                          (i < freeFramePoolObj->numDynamicHdrBufs);
                          i++) {
                        if (    portablePtr
                             == freeFramePoolObj->dynHdrBufs[i]) {
                            /* Find out the original frame to which the given
                             * duped frame is referring to.
                             */
                            refIndex =
                                     freeFramePoolObj->dynHdrBuf_refCntIndex[i];
                            lFrameIndex = refIndex;
                            /* Check if this Header is pointing to dynamic frame
                             * or static frame and increment ref count of static
                             * frame or dynamic frame.
                             */
                            if (   (freeFramePoolObj->
                                    dynHdrBuf_refCntIndexType[i])
                                ==  (FrameQBufMgr_ShMem_STATIC_FRAME)) {
                                freeFramePoolObj->frmRefCnt[refIndex] +=
                                                                 numDupedFrames;
                                indexType = FrameQBufMgr_ShMem_STATIC_FRAME;

                            }
                            else {
                                /* This header buffers is actually duped frame for frame
                                 * whcih belongs dynamic frames of this queue
                                 */
                                freeFramePoolObj->dynFrmRefCnt[refIndex]
                                                              += numDupedFrames;
                                indexType = FrameQBufMgr_ShMem_DYNAMIC_FRAME;

                            }

                        }
                    }
                } /*if (bDuped == FALSE) {*/
            }
            /* Now Allocate  duped headers  from the quque and copy the contents
             * of the orginal frame to duped frames  and  return
             */
            if (status >= 0) {
                for (k = 0;
                     (  (k < freeFramePoolObj->numFrmHdrBufs)
                      &&(ldupedFrames < numDupedFrames)
                      &&(freeFramePoolObj->freeFrmHdrBufs > 0));
                     k++) {
                    if (freeFramePoolObj->hdrBufIsUsed[k] != TRUE){
                        /* Found free header */
                        freeFramePoolObj->hdrBufIsUsed[k] = TRUE;

                        freeFramePoolObj->hdrBuf_refCntIndex[k] =
                                                                   lFrameIndex;
                        /* Set if it is pointing to static frame or
                         * dynamic frame.
                         */
                        freeFramePoolObj->hdrBuf_refCntIndexType[k] =
                                                                     indexType;
                        freeFramePoolObj->freeFrmHdrBufs--;
                        portablePtr = (UInt32)
                          (   freeFramePoolObj->hdrBufStartAddrp
                            + (  (   (  (freeFramePoolObj->hdrBufEndAddrp)
                                       -(freeFramePoolObj->hdrBufStartAddrp))
                                   / (freeFramePoolObj->numFrmHdrBufs))
                                * k) );

                        lframe = SyslinkMemMgr_translate (obj->memMgrAllocater_hdrBufs,
                                                   (Ptr)portablePtr,
                                                   SyslinkMemMgr_AddrType_Portable,
                                                   SyslinkMemMgr_AddrType_Virtual);
                        /* Copy  given frame header contens to this frame*/
                        memcpy(lframe,
                               framePtr[lFrmCnt],
                               framePtr[lFrmCnt]->headerSize);
                        dupedFramePtr[lFrmCnt][ldupedFrames] = lframe;
                        ldupedFrames++;
                    }

                }/*for (k = 0;  */

                if (ldupedFrames < numDupedFrames) {
                    /* If still  need to allocate */
                    for (k = 0;
                        (  (k < freeFramePoolObj->numDynamicHdrBufs)
                         &&(ldupedFrames < numDupedFrames)
                         &&(freeFramePoolObj->freeDynamicHdrBufs > 0));
                        k++) {
                        if (freeFramePoolObj->dynHdrBufIsUsed[k] != TRUE){
                            /* Found free header */
                            freeFramePoolObj->dynHdrBufIsUsed[k] = TRUE;

                            freeFramePoolObj->dynHdrBuf_refCntIndex[k] =
                                                                    lFrameIndex;
                            /* Set if it is pointing to static frame or
                             * dynamic frame.
                             */
                            freeFramePoolObj->dynHdrBuf_refCntIndexType[k]
                                                                   = indexType;
                            freeFramePoolObj->freeDynamicHdrBufs--;
                            portablePtr = (UInt32)
                                          freeFramePoolObj->dynHdrBufs[k];

                            lframe = SyslinkMemMgr_translate (
                                                   obj->memMgrAllocater_hdrBufs,
                                                   (Ptr)portablePtr,
                                                   SyslinkMemMgr_AddrType_Portable,
                                                   SyslinkMemMgr_AddrType_Virtual);
                            /* Copy  given frame header contens to this frame*/
                            memcpy(lframe,
                                   framePtr[lFrmCnt],
                                   framePtr[lFrmCnt]->headerSize);
                            dupedFramePtr[lFrmCnt][ldupedFrames] = lframe;
                            ldupedFrames++;
                        }

                    }/*for (k = 0;  */

                }/*if (ldupedFrames < numDupedFrames) */
            }
            /* Writeback contents of the frame pool */
            if (obj->ctrlStructCacheFlag == TRUE) {
                Cache_wbInv ((Ptr)freeFramePoolObj,
                             sizeof (FrameQBufMgr_ShMem_FreeFramePoolObj),
                             Cache_Type_ALL,
                             TRUE);
            }
        }

    }

    /* Release lock */
    GateMP_leave ((GateMP_Handle)obj->gate, key);

    return (status);
}

/*
 *  ======== FrameQBufMgr_ShMem_writeBack ========
 * Write back the contents  of a frame including frame buffers.
 *
 * NOTE: frame address and frame buffer address in the frame should be in
 *       virtual format.
 */
Int32
FrameQBufMgr_ShMem_writeBack(FrameQBufMgr_ShMem_Object *obj,
                             FrameQBufMgr_ShMem_Frame   frame)
{
    Int32   status = FrameQBufMgr_S_SUCCESS;
    FrameQBufMgr_ShMem_FrameBufInfo   *frameBufInfo;
    Ptr     addr;
    UInt32  i;

    Assert_isTrue ((NULL != obj), FrameQBufMgr_ShMem_A_invalidParams);

    Assert_isTrue ((NULL != frame), FrameQBufMgr_ShMem_A_invalidParams);

    frameBufInfo = (FrameQBufMgr_ShMem_FrameBufInfo*) &(frame->frameBufInfo[0]);

    for (i = 0; i < frame->numFrameBuffers; i++) {
        /* Write  back the buffer if cache flag  and cpu access flag is enabled.
         */
        if (   (obj->frmBufCacheFlag[i] == TRUE)
            && (obj->bufCpuAccessFlag[i] == TRUE)) {

            /* Write back only valid contents in frame buffer */
            addr = (Ptr)(   (UInt32)(frameBufInfo[i].bufPtr)
                          + (UInt32)(frameBufInfo[i].startOffset));
            Cache_wbInv(addr,
                        frameBufInfo[i].validSize,
                        Cache_Type_ALL,
                        TRUE);
            if(obj->bufInterfaceType == FrameQBufMgr_BUFINTERFACE_SHAREDMEM){
                frameBufInfo[i].bufPtr = (UInt32)SyslinkMemMgr_translate (
                                                   obj->memMgrAllocater_hdrBufs,
                                                   (Ptr)frameBufInfo[i].bufPtr,
                                                       SyslinkMemMgr_AddrType_Virtual,
                                                       SyslinkMemMgr_AddrType_Portable);
            }
            if(obj->bufInterfaceType == FrameQBufMgr_BUFINTERFACE_TILERMEM){
                /* Currently no translation is required right now*/
            }
        }
    }

    /* Write  back frame  if cache flag  is enabled.
     */
    if (obj->frmHdrBufCacheFlag == TRUE) {
        Cache_wbInv(frame,
                    frame->headerSize,
                    Cache_Type_ALL,
                    TRUE);
    }

    return (status);
}

/*
 *  ======== FrameQBufMgr_ShMem_invalidate ========
 * Invalidate the contents  of a frame including frame buffers.
 *
 *
 * frame  address in virtual format.
 * It converts the internal portable frame buffer pointers to virtual format in
 * the frame.
 */
Int32
FrameQBufMgr_ShMem_invalidate(FrameQBufMgr_ShMem_Object *obj,
                              FrameQBufMgr_ShMem_Frame frame)
{
    Int32  status = FrameQBufMgr_S_SUCCESS;
    FrameQBufMgr_ShMem_FrameBufInfo *frameBufInfo;
    Ptr    addr;
    UInt32 i;

    Assert_isTrue ((NULL != obj), FrameQBufMgr_ShMem_A_invalidParams);

    Assert_isTrue ((NULL != frame), FrameQBufMgr_ShMem_A_invalidParams);

    frameBufInfo = (FrameQBufMgr_ShMem_FrameBufInfo*) &(frame->frameBufInfo[0]);

    /* Write  back frame  if cache flag  is enabled.
     */
    if (obj->frmHdrBufCacheFlag == TRUE) {
        /* Do invalidation first one cache line */
        Cache_inv(frame,
                  FrameQBufMgr_CACHE_LINESIZE,
                  Cache_Type_ALL,
                  TRUE);
        Cache_inv(frame,
                  frame->headerSize,
                  Cache_Type_ALL,
                  TRUE);
    }

    for (i = 0; i < frame->numFrameBuffers; i++) {
        /* Write  back the buffer if cache flag  and cpu access flag is enabled.
         */
        if (   (obj->frmBufCacheFlag[i] == TRUE)
            && (obj->bufCpuAccessFlag[i] == TRUE)) {
            if(obj->bufInterfaceType == FrameQBufMgr_BUFINTERFACE_TILERMEM){
                frameBufInfo[i].bufPtr = (UInt32)SyslinkMemMgr_translate (
                                     obj->memMgrAllocater_hdrBufs,
                                    (Ptr)frameBufInfo[i].bufPtr,
                                         SyslinkMemMgr_AddrType_Portable,
                                         SyslinkMemMgr_AddrType_Virtual);

                /* Write back only valid contents in frame buffer */
                addr = (Ptr)(   (UInt32)(frameBufInfo[i].bufPtr)
                              + (UInt32)(frameBufInfo[i].startOffset));
                Cache_inv(addr,
                          frameBufInfo[i].validSize,
                          Cache_Type_ALL,
                          TRUE);
            }
            if(obj->bufInterfaceType == FrameQBufMgr_BUFINTERFACE_TILERMEM){
                /* Currently no translation is required right now*/
            }
        }
    }

    return (status);
}

/*
 *  ======== FrameQBufMgr_ShMem_writeBackBuf ========
 * Write back the contents  of buffer.  API treates the passed as frame Header.
 */
Int32
FrameQBufMgr_ShMem_writeBackHeaderBuf(FrameQBufMgr_ShMem_Object *obj,
                                      Ptr headerBuf)
{

    Int32  status = FrameQBufMgr_S_SUCCESS;
    FrameQBufMgr_ShMem_Frame  frame = (FrameQBufMgr_ShMem_Frame)headerBuf;

    Assert_isTrue ((NULL != obj), FrameQBufMgr_ShMem_A_invalidParams);

    Assert_isTrue ((NULL != frame), FrameQBufMgr_ShMem_A_invalidParams);

    if (obj->frmHdrBufCacheFlag == TRUE) {
        Cache_wbInv(frame,
                    frame->headerSize,
                    Cache_Type_ALL,
                    TRUE);
    }
    return (status);
}

/*
 *  ======== FrameQBufMgr_ShMem_writeBackBuf ========
 * Write back the contents  of buffer . Flags denotes whether it is header buffer
 * or frame  buffer.
 */
Int32
FrameQBufMgr_ShMem_invalidateHeaderBuf(FrameQBufMgr_ShMem_Object *obj,
                                       Ptr headerBuf)
{

    Int32  status = FrameQBufMgr_S_SUCCESS;
    FrameQBufMgr_ShMem_Frame  frame = (FrameQBufMgr_ShMem_Frame)headerBuf;

    Assert_isTrue ((NULL != obj), FrameQBufMgr_ShMem_A_invalidParams);

    Assert_isTrue ((NULL != frame), FrameQBufMgr_ShMem_A_invalidParams);

    if (obj->frmHdrBufCacheFlag == TRUE) {
        /* Do invalidation first one cache line */
        Cache_wbInv(frame,
                    FrameQBufMgr_ShMem_CACHE_LINESIZE,
                    Cache_Type_ALL,
                    TRUE);
        Cache_wbInv(frame,
                    frame->headerSize,
                    Cache_Type_ALL,
                    TRUE);
    }

    return (status);
}

/*
 *  ======== FrameQBufMgr_ShMem_writeBackFrameBuf ========
 * Write back the contents  of buffer . Passed buffer is treated as frame  buffer.
 * bufNumInFrame denotes  buffer number in the frame.This is  to identify
 * the cache flags and cpu access flags for the  buffer.
 */
Int32
FrameQBufMgr_ShMem_writeBackFrameBuf(FrameQBufMgr_ShMem_Object *obj,
                                     Ptr                        frameBuf,
                                     UInt32                     size,
                                     UInt8                      bufIndexInFrame)
{
    Int32         status = FrameQBufMgr_S_SUCCESS;

    Assert_isTrue ((NULL != obj), FrameQBufMgr_ShMem_A_invalidParams);

    Assert_isTrue ((NULL != frameBuf), FrameQBufMgr_ShMem_A_invalidParams);

    Assert_isTrue (( 8 > bufIndexInFrame), FrameQBufMgr_ShMem_A_invalidParams);

    /* Find out the buffer type from the flags */
    if (   (obj->frmBufCacheFlag[bufIndexInFrame] == TRUE)
        && (obj->bufCpuAccessFlag[bufIndexInFrame] == TRUE)) {
        Cache_wbInv(frameBuf,
                    size,
                    Cache_Type_ALL,
                    TRUE);
    }

    return (status);
}

/*
 *  ======== FrameQBufMgr_ShMem_invalidateFrameBuf ========
 * Invalidate the contents  of frame buffer .
 *
 * bufNumInFrame denotes which buffer it is  to identify the cache flags and
 * cpu access flags for the  buffer. This is the sequence number of frame buffer
 * in the frame.
 */
Int32
FrameQBufMgr_ShMem_invalidateFrameBuf(FrameQBufMgr_ShMem_Object *obj,
                                      Ptr                       frameBuf,
                                      UInt32                    size,
                                      UInt8                     bufIndexInFrame)
{

    Int32  status = FrameQBufMgr_S_SUCCESS;

    Assert_isTrue ((NULL != obj), FrameQBufMgr_ShMem_A_invalidParams);

    Assert_isTrue ((NULL != frameBuf), FrameQBufMgr_ShMem_A_invalidParams);

    Assert_isTrue (( 8 > bufIndexInFrame), FrameQBufMgr_ShMem_A_invalidParams);

    /* Find out the buffer type from the flags */
    if (   (obj->frmBufCacheFlag[bufIndexInFrame] == TRUE)
        && (obj->bufCpuAccessFlag[bufIndexInFrame] == TRUE)) {
        Cache_inv(frameBuf,
                  size,
                  Cache_Type_ALL,
                  TRUE);
    }

    return (status);
}

/*
 *  ======== FrameQBufMgr_ShMem_translateAddr ========
 * Function to translate address between different types.
 * see also FrameQBufMgr_ShMem_AddrType
 */
Int32
FrameQBufMgr_ShMem_translateAddr (FrameQBufMgr_ShMem_Object     *obj,
                                  Ptr *                          dstAddr,
                                  FrameQBufMgr_ShMem_AddrType    dstAddrType,
                                  Ptr                            srcAddr,
                                  FrameQBufMgr_ShMem_AddrType    srcAddrType,
                                  FrameQBufMgr_ShMem_BufType     bufType)
{
    Int32         status = FrameQBufMgr_S_SUCCESS;
    SyslinkMemMgr_Handle memMgrHandle = NULL;
    Ptr           addr;
    SyslinkMemMgr_AddrType        sAddrType;
    SyslinkMemMgr_AddrType        dAddrType;

    Assert_isTrue ((NULL != obj), FrameQBufMgr_ShMem_A_invalidParams);

    Assert_isTrue ((NULL != dstAddr), FrameQBufMgr_ShMem_A_invalidParams);

    Assert_isTrue ((NULL != srcAddr), FrameQBufMgr_ShMem_A_invalidParams);

    switch (bufType)
    {
        case FrameQBufMgr_ShMem_BUF_FRAMEHEADER:
        {
            memMgrHandle = obj->memMgrAllocater_hdrBufs;
        }
        break;

        case FrameQBufMgr_ShMem_BUF_FRAMEBUF:
        {
            memMgrHandle = obj->memMgrAllocater_frmBufs;
        }
        break;

        default:
            status = FrameQBufMgr_E_FAIL;
    }
    if ((status >= 0)) {
        if (dstAddrType == FrameQBufMgr_ShMem_AddrType_Virtual) {
            dAddrType = SyslinkMemMgr_AddrType_Virtual;
        }
        else {
            dAddrType = SyslinkMemMgr_AddrType_Portable;
        }

        if (srcAddrType == FrameQBufMgr_ShMem_AddrType_Virtual) {
            sAddrType = SyslinkMemMgr_AddrType_Virtual;
        }
        else {
            sAddrType = SyslinkMemMgr_AddrType_Portable;
        }

        addr = SyslinkMemMgr_translate (memMgrHandle,
                                        (Ptr)srcAddr,
                                        sAddrType,
                                        dAddrType);
        if (addr ==NULL) {
            status = FrameQBufMgr_E_FAIL;
            *dstAddr = (Ptr)NULL;
        }
        else {
            *dstAddr = (Ptr)addr;
        }
    }

    return (status);
}

/*
 *  ======== FrameQBufMgr_ShMem_getId ========
 * Function to get FrameQbufMgr Id.
 */
UInt32
FrameQBufMgr_ShMem_getId (FrameQBufMgr_ShMem_Object* obj)
{
    Assert_isTrue ((NULL != obj), FrameQBufMgr_ShMem_A_invalidParams);

    return (obj->instId);
}

/*
 *  ======== FrameQBufMgr_ShMem_getHandle ========
 *  Internal API to return the FrameQbufMgr Id.
 */
Ptr
FrameQBufMgr_ShMem_getHandle (UInt32 Id)
{
    FrameQBufMgr_ShMem_Handle             handle  = NULL;
    FrameQBufMgr_ShMem_Handle             nextHandle = NULL;
    UInt32                           i;
    UInt32                           key;
    FrameQBufMgr_ShMem_Object             *obj;
    UInt16                          instId;
    UInt16                          instNo;
    UInt16                          creProcId;

    instId = Id & 0xFFFF;
    /* Find out the creator procid */
    creProcId = (  (instId >> FRAMEQBUFMGR_CREATORPROCID_BITOFFSET)
                  & FRAMEQBUFMGR_CREATORPROCID_MASK);
    Assert_isTrue ((creProcId < ti_sdo_utils_MultiProc_numProcessors),
                     FrameQBufMgr_ShMem_A_InvalidCreatorProcId);

    /* Get the instance no of the instance */
    instNo = ( (instId >> FRAMEQBUFMGR_INSTNO_BITOFFSET)
              & FRAMEQBUFMGR_INSTNO_MASK);
    Assert_isTrue ((instNo < FrameQBufMgr_MAXINSTANCES),
                     FrameQBufMgr_ShMem_A_InvalidInstNo);

    if (instNo < FrameQBufMgr_MAXINSTANCES) {

        key = IGateProvider_enter (FrameQBufMgr_ShMem_module->gate );
        for ( handle = FrameQBufMgr_ShMem_Object_first();
              handle != NULL;
              handle = nextHandle) {
              obj =  handle;
               if (obj->instId == instId) {
                    /* Found the handle. return it */
                    break;
               }
              nextHandle = FrameQBufMgr_ShMem_Object_next(handle) ;
        }

        if (handle == NULL ) {
            for (i = 0; i < FrameQBufMgr_ShMem_Object_count(); i++) {
               handle = FrameQBufMgr_ShMem_Object_get(NULL, i);
               obj =  handle;
               if (obj->instId == instId) {
                    /* Found the  handle. return it */
                    break;
               }
            }
        }
        IGateProvider_leave (FrameQBufMgr_ShMem_module->gate, key );

    }/*if (instNo < FrameQBufMgr_MAXINSTANCES) */

    return ((Ptr)handle);
}

/*
 *  ======== FrameQBufMgr_ShMem_getNumFreeFrames ========
 *  Function to get the number of free frames available in
 *  FrameQbufMgr's FreeFramePool 0. The returned number free frame may not
 *  exist in the frameQbufMgr  after this call.This is because
 *  if any allocs are done by the other FrameQ writers on the
 *  same FrameQBufMgr.If frees are not other frameQ writers the free Frames
 *  would be even more than the value that we have got.
 *
 */
Int32
FrameQBufMgr_ShMem_getNumFreeFrames (FrameQBufMgr_ShMem_Object *obj,
                                     UInt32                    * numFreeFrames)
{
    Int32                              status           = FrameQBufMgr_S_SUCCESS;
    UInt32                             freeFramePoolIndex = 0;
    volatile FrameQBufMgr_ShMem_FreeFramePoolObj     *freeFramePoolObj;
    UInt32                             key;

    Assert_isTrue ((NULL != obj),
                   FrameQBufMgr_ShMem_A_invalidParams);

    Assert_isTrue ((NULL != numFreeFrames),
                   FrameQBufMgr_ShMem_A_invalidParams);

    key = GateMP_enter ((GateMP_Handle)obj->gate);

    /* Alloc allocates frame from the free Frame Pool 0*/
    freeFramePoolIndex = 0;
    freeFramePoolObj =  obj->freeFramePool[freeFramePoolIndex];
    if (obj->ctrlStructCacheFlag == TRUE) {
        Cache_inv ((Ptr)freeFramePoolObj,
                   sizeof(FrameQBufMgr_ShMem_FreeFramePoolObj),
                   Cache_Type_ALL,
                   TRUE);
    }

    * numFreeFrames = (  freeFramePoolObj->freeFrms
                       + freeFramePoolObj->freeDynamicFrames);
    GateMP_leave ((GateMP_Handle)obj->gate, key);

    return (status);
}

/*
 *  ======== FrameQBufMgr_ShMem_getvNumFreeFrames ========
 * Function to get the number of free frames available in the plugged in
 * FrameQbufMgr.
 */
Int32
FrameQBufMgr_ShMem_getvNumFreeFrames (
                                FrameQBufMgr_ShMem_Object   * obj,
                                UInt32                        numFreeFrames[],
                                UInt8                         freeFramePoolNo[],
                                UInt8                         numFreeFramePools)
{
    Int32                              status           = FrameQBufMgr_S_SUCCESS;
    volatile FrameQBufMgr_ShMem_Attrs  *pAttrs          = NULL;
    UInt32                             freeFramePoolIndex;
    volatile FrameQBufMgr_ShMem_FreeFramePoolObj     *freeFramePoolObj;
    UInt32                             i;
    UInt32                             key;

    Assert_isTrue ((NULL != obj),
                   FrameQBufMgr_ShMem_A_invalidParams);

    Assert_isTrue ((NULL != numFreeFrames),
                   FrameQBufMgr_ShMem_A_invalidParams);

    Assert_isTrue ((NULL != freeFramePoolNo),
                   FrameQBufMgr_ShMem_A_invalidParams);

    pAttrs = obj->attrs;

    /* Acquire lock */
    key = GateMP_enter ((GateMP_Handle)obj->gate);

    /*  Invalidate control if required */
    if (obj->ctrlStructCacheFlag == TRUE) {
        Cache_inv((Ptr)pAttrs,
                  sizeof (FrameQBufMgr_ShMem_Attrs),
                  Cache_Type_ALL,
                  TRUE);
    }

    for (i = 0; i < numFreeFramePools; i++) {
        freeFramePoolIndex = freeFramePoolNo[i];
        /* Assert if pool no given is more than the configured pools*/
        Assert_isTrue ((freeFramePoolIndex < pAttrs->numFreeFramePools),
                        FrameQBufMgr_ShMem_A_invalidParams);
        freeFramePoolObj =  obj->freeFramePool[freeFramePoolIndex];
        if (obj->ctrlStructCacheFlag == TRUE) {
            Cache_inv ((Ptr)freeFramePoolObj,
                       sizeof(FrameQBufMgr_ShMem_FreeFramePoolObj),
                       Cache_Type_ALL,
                       TRUE);
        }
        numFreeFrames[i] = (  freeFramePoolObj->freeFrms
                            + freeFramePoolObj->freeDynamicFrames);
    }
    /* Release lock */
    GateMP_leave ((GateMP_Handle)obj->gate, key);

    return (status);
}

/*
 *  ======== FrameQBufMgr_ShMem_isCacheEnabledForFrameBuf ========
 * Function to find out if cache is enabled for header buffers.
 */
Bool FrameQBufMgr_ShMem_isCacheEnabledForHeaderBuf(FrameQBufMgr_ShMem_Object * obj)
{
    Assert_isTrue ((NULL != obj),
                   FrameQBufMgr_A_invalidParams);

    return (obj->frmHdrBufCacheFlag);
}

/*
 *  ======== FrameQBufMgr_ShMem_isCacheEnabledForFrameBuf ========
 * Function to find out if cache is enabled for frame buffers.
 */
Bool FrameQBufMgr_ShMem_isCacheEnabledForFrameBuf(FrameQBufMgr_ShMem_Object * obj,
                                                  UInt8 frameBufIndex)
{
    Assert_isTrue ((NULL != obj),
                   FrameQBufMgr_A_invalidParams);

    Assert_isTrue ((FrameQBufMgr_MAX_FRAMEBUFS > frameBufIndex),
                   FrameQBufMgr_A_invalidParams);

    return (obj->frmBufCacheFlag[frameBufIndex]);
}

/*
 *  ======== FrameQBufMgr_ShMem_control ========
 *  @brief  Provides a hook to perform implementation dependent operation.
 *
 */
Int32
FrameQBufMgr_ShMem_control (FrameQBufMgr_ShMem_Object *obj,
                            Int32                      cmd,
                            Ptr                        arg)
{
    Int32   status = FrameQBufMgr_S_SUCCESS;

    Assert_isTrue ((NULL != obj),
                   FrameQBufMgr_ShMem_A_invalidParams);

    status = FrameQBufMgr_E_NOTIMPLEMENTED;

    return (status);
}

/******************************* Internal functions ***************************/
/*
 *  ======== FrameQBufMgr_ShMem_Module_startup ========
 */
Int FrameQBufMgr_ShMem_Module_startup(Int phase)
{
    Int32                        key;
    GateSwi_Handle               swiGate;
    Error_Block                  eb;

    if (FrameQBufMgr_ShMem_module->nameServer != NULL) {
        if (!ti_sdo_utils_NameServer_Module_startupDone()) {
            return (Startup_NOTDONE);
        }
    }
    /* Check the  implementation specific module setups */
    if (!FrameQBufMgr_Module_startupDone()) {
        return (Startup_NOTDONE);
    }

    Error_init(&eb);

    key = Gate_enterSystem();

    FrameQBufMgr_ShMem_module->refCount++;

    if(FrameQBufMgr_ShMem_module->refCount > 1) {
        Gate_leaveSystem(key);
        return (Startup_DONE);
    }
    else {
        /*Open the Name server created by FrameQ front end*/
        if (   (FrameQBufMgr_ShMem_module->cfg.useNameServer == TRUE)
                && (FrameQBufMgr_ShMem_module->nameServer == NULL)) {
            /* Open  the Name Server instance  created by Frond End FrameQBufMgr
             * module if not  assigned statically.
             */
//            FrameQBufMgr_ShMem_module->nameServer = NameServer_getHandle(
//                                                   FrameQBufMgr_NAMESERVERNAME);
            FrameQBufMgr_ShMem_module->nameServer = gFrameQBufMgr_nsHandle;
            if(NULL == FrameQBufMgr_ShMem_module->nameServer) {

                /* Decrement the reference count */
                FrameQBufMgr_ShMem_module->refCount--;
                Gate_leaveSystem(key);

                return (Startup_NOTDONE);
            }
        }
    }/* if(FrameQBufMgr_ShMem_module->refCount > 1) { */

    /* Release lock */
    Gate_leaveSystem(key);

    if (FrameQBufMgr_ShMem_module->gate == NULL) {
        /*  Create local SWI gate  for module wide protection
         * if not created statically.
         */
        swiGate = GateSwi_create(NULL, &eb);
        if (swiGate != NULL) {
            FrameQBufMgr_ShMem_module->gate =
                                      GateSwi_Handle_upCast(swiGate);
            FrameQBufMgr_ShMem_module->gateType =
                                      FrameQBufMgr_ShMem_DYNAMIC_CREATE;
        }
        else {
            return (Startup_NOTDONE);
        }
    }

    return (Startup_DONE);
}

/*
 *  ======== _FrameQBufMgr_ShMem_create ========
 *  API to create/open the instance.Called internally by
 *  the FrameQBufMgr_create call.
 */
static inline
Int32
_FrameQBufMgr_ShMem_create(FrameQBufMgr_ShMem_Object       * obj,
                           const ti_syslink_ipc_rtos_shMem_FrameQBufMgr_ShMem_Params * params,
                           Error_Block *eb)
{
    Int32                         status         = FrameQBufMgr_S_SUCCESS;
    UInt32                        i              = 0;
    volatile FrameQBufMgr_ShMem_Attrs * pAttrs   = NULL;
    UInt32                        addrOffset     = 0;
    UInt32                        virtAddr       = 0;
    UInt16                        instId         = 0;
    UInt                          align          = 4;
    Int32                         tmpStatus;
    ClientNotifyMgr_Params        clientMgrParams;
    ClientNotifyMgr_Handle        cliMgrHandle;
    UInt32                        regionId;
    SharedRegion_SRPtr            sharedShmBase;
    GateMP_Params                 gateParams;
    GateMP_Handle                 gateHandle;
    UInt32                        key;
    UInt8                         nameLen;
    FrameQBufMgr_NameServerEntry  nameserverEntry;
    FrameQBufMgr_NameServerEntry *pnameserverEntry;
    UInt32                        tempHdrChunk;
    UInt32                        tempBufChunk;
    UInt32                        j;
    UInt8                         instNo;
    UInt32                        gateMPsharedAddrSize;
    FrameQBufMgr_ShMem_Params     sparams;

    Assert_isTrue ((NULL != obj), FrameQBufMgr_ShMem_A_invalidParams);

    Assert_isTrue ((NULL != params), FrameQBufMgr_ShMem_A_invalidParams);

    /* Assert if shared addr  provided is null  and region id provide is
     * invalid.
     */
    Assert_isTrue (    (params->sharedAddr != NULL)
                    || (params->regionId != SharedRegion_INVALIDREGIONID),
                    FrameQBufMgr_ShMem_A_invalidParams);

    /* Assert if FreeFramePools are zero */
    Assert_isTrue ((params->numFreeFramePools != 0),
                    FrameQBufMgr_ShMem_A_invalidParams);

    Assert_isTrue ((params->numFreeFramePools <= FrameQBufMgr_ShMem_MAX_POOLS),
                    FrameQBufMgr_ShMem_A_invalidParams);

    FrameQBufMgr_ShMem_getSharedParams(&sparams, params);
    /* Assert if shared addr size is less than  shared memory requirement */
    FrameQBufMgr_ShMem_sharedMemReq(&sparams,
                                    &tempHdrChunk,
                                    &tempBufChunk);
//    Assert_isTrue ( (   (params->sharedAddr != NULL)
//                    &&     params->sharedAddrSize
//                         >= FrameQBufMgr_ShMem_sharedMemReq(&sparams,
//                                                            &tempHdrChunk,
//                                                            &tempBufChunk)),
//                    FrameQBufMgr_ShMem_A_invalidParams);

    /* Validate the params */
    /* Check if Buffer address is given to create buffer pools*/
    if (params->sharedAddrHdrBuf != NULL) {
        if (params->sharedAddrHdrBufSize < tempHdrChunk) {
            status = FrameQBufMgr_E_INVALIDARG;
            Error_raise(eb,
                        FrameQBufMgr_E_InvalidParams,
                        NULL,
                        0);

        }
        else {
            if ((  ((SizeT)params->sharedAddrHdrBuf)
                 & (align-1)) != 0) {
                status = FrameQBufMgr_E_INVALIDARG;
                Error_raise(eb,
                            FrameQBufMgr_E_InvalidParams,
                            NULL,
                            0);
            }
        }
    }

    if (params->sharedAddrFrmBuf != NULL) {
        if (params->sharedAddrFrmBufSize < tempBufChunk) {
            status = FrameQBufMgr_E_INVALIDARG;
            Error_raise(eb,
                        FrameQBufMgr_E_InvalidParams,
                        NULL,
                        0);
        }
        else {
            if ((  ((SizeT)params->sharedAddrFrmBuf)
                 & (align-1)) != 0) {
                status = FrameQBufMgr_E_INVALIDARG;
                Error_raise(eb,
                            FrameQBufMgr_E_InvalidParams,
                            NULL,
                            0);
            }
        }
    }

    if (status >= 0) {
        /*
         * Ensure buffer Sizes specified in the bufSizes array
         * is multiple of align.
         */
        for ( i= 0; i < params->numFreeFramePools; i++ ) {
            /* Ensure bufSizes[i] is a multiple of cache align.  */
            if ((((SizeT)params->frameHeaderBufSize[i])
                 % FrameQBufMgr_ShMem_CACHE_LINESIZE) != 0) {
                status = FrameQBufMgr_E_INVALIDARG;
                Error_raise(eb,
                            FrameQBufMgr_ShMem_E_CacheAlign,
                            NULL,
                            0);
                break;
            }
        }
    }

    if (status >= 0) {
        /* Ensure buffer Sizes specified in the bufSizes array
         * is multiple of align.
         */
        for ( i = 0;
             (i < params->numFreeFramePools) && (status >= 0); i++ ) {
            for (j = 0; j < params->numFrameBufsInFrame[i]; j++) {
                /* Ensure bufSizes[i] is a multiple of cache align.*/
                if ((((SizeT)params->frameBufParams[i][j].size)
                     % FrameQBufMgr_ShMem_CACHE_LINESIZE) != 0) {
                    status = FrameQBufMgr_E_INVALIDARG;
                    Error_raise(eb,
                                FrameQBufMgr_ShMem_E_CacheAlign,
                                NULL,
                                0);
                    break;
                }
            }
        }
    }
    if (status >= 0) {
        for ( i= 0; i < params->numFreeFramePools; i++ ) {
            Assert_isTrue((params->numFreeFramesInPool[i] <= FrameQBufMgr_ShMem_POOL_MAXFRAMES),
                           FrameQBufMgr_ShMem_A_invalidParams);
            Assert_isTrue((params->numFreeHeaderBufsInPool[i] <= FrameQBufMgr_ShMem_POOL_MAXHDRS),
               FrameQBufMgr_ShMem_A_invalidParams);
        }
    }

    if (status >= 0) {
        /*Call FrameQBufMgr API to get teh unique instance number */
        status = FrameQBufMgr_getInstNo(&instNo) ;
        if (status >= 0) {
            /* Found the  empty slot. This is the unique
             * instnace no for this instance on this processor
             * By appending creator proc id we get the unique
             * instance id for this instance in the entire system
             */
            instId = ( (   MultiProc_self()
                         & FRAMEQBUFMGR_CREATORPROCID_MASK)
                       << FRAMEQBUFMGR_CREATORPROCID_BITOFFSET)
                    + (   (instNo & FRAMEQBUFMGR_INSTNO_MASK)
                       << FRAMEQBUFMGR_INSTNO_BITOFFSET);

            /* Set the Id to default */
            obj->instId = (UInt32)-1;

            /* The memory pointed to by sharedAddr must be in shared memory and
             * the address must be mapped in all processors which require access
             * to this instance.
             */
            if (params->sharedAddr == NULL) {
                /* Allocate memory for controlStruct as user has not provided it
                 */
                /* Creating using a shared region ID and name */
                obj->regionId = params->regionId;
                obj->ctrlStructCacheFlag =
                                    SharedRegion_isCacheEnabled(obj->regionId);

                /* Need to allocate from the heap */
                obj->allocSize = FrameQBufMgr_ShMem_sharedMemReq(
                                             (FrameQBufMgr_ShMem_Params*)&sparams,
                                             &tempHdrChunk,
                                             &tempBufChunk);
                /* HeapMemMP will do the alignment */
                virtAddr = (UInt32)Memory_alloc(
                                        SharedRegion_getHeap(obj->regionId),
                                        obj->allocSize,
                                        0,
                                        eb);
            }
            else {
                /* Creating using sharedAddr */
                obj->regionId = SharedRegion_getId(params->sharedAddr);

                /* Assert that the buffer is in a valid shared region */
                Assert_isTrue(obj->regionId != SharedRegion_INVALIDREGIONID,
                              Ipc_A_addrNotInSharedRegion);

                /* Assert that sharedAddr is cache aligned */
                Assert_isTrue(
                            ((UInt32)params->sharedAddr %
                             SharedRegion_getCacheLineSize(obj->regionId) == 0),
                        Ipc_A_addrNotCacheAligned);

                /* set object's cacheEnabled, objType, and attrs  */
                obj->ctrlStructCacheFlag =
                                 SharedRegion_isCacheEnabled(obj->regionId);
                virtAddr = (UInt32)params->sharedAddr;
            }

            Assert_isTrue ((NULL != virtAddr),
                            FrameQBufMgr_ShMem_A_nullPointer);

            obj->pnameserverEntry = (FrameQBufMgr_NameServerEntry *)virtAddr;
            /*
             * Populate nameserver entry params that will be stored in
             * nameserver as well as at the start of the shared Addr .
             */
            nameserverEntry.ctrlInterfaceType =
                                   FrameQBufMgr_ShMem_INTERFACE_SHAREDMEM;
            nameserverEntry.headerInterfaceType =
                                   FrameQBufMgr_ShMem_HDRINTERFACE_SHAREDMEM;
            nameserverEntry.bufInterfaceType =
                                   params->bufInterfaceType;
            regionId = SharedRegion_getId (
                                          (Ptr)virtAddr);
            sharedShmBase = SharedRegion_getSRPtr ((Ptr)virtAddr, regionId);

            nameserverEntry.instParams.shMemParams.sharedAddr = (Ptr)
                                                              sharedShmBase;
            if (params->sharedAddr != NULL) {
                nameserverEntry.instParams.shMemParams.sharedAddrSize =
                                                         params->sharedAddrSize;
            }
            else {
                nameserverEntry.instParams.shMemParams.sharedAddrSize =
                                                         obj->allocSize;
            }

            /* Pass region id  to be used to be used
             * internally to allocate headers.
             * TODO:In case of Tiler what should  we pass here.
             */
            nameserverEntry.instParams.shMemParams.hdrBuf_regionId =
                                                        params->hdrBuf_regionId;

            /* Pass region id  to be used
             * internally to allocate  frame buffers.
             * TODO:In case of Tiler what should  we pass here.
             */
            nameserverEntry.instParams.shMemParams.frmBuf_regionId =
                                                        params->frmBuf_regionId;

            /*******************************************************************
             * Create instance gate based on the parameters if user has not
             * provided gate.
             *******************************************************************
             */
            if (params->gate != NULL) {
                obj->gate = params->gate;
                nameserverEntry.instParams.shMemParams.instGateMPAddr =
                           (Ptr) ti_sdo_ipc_GateMP_getSharedAddr(obj->gate);
                obj->creGate = FALSE;
            }
            else {
                if (FrameQBufMgr_ShMem_usedefaultgate == TRUE) {
                    obj->creGate = FALSE;
                    /* Check if FrameQ module has any default gate set. if not use
                     * GateMP's default gate.
                     */
                    if (FrameQBufMgr_ShMem_module->defaultInstGate != NULL) {
                        gateHandle = FrameQBufMgr_ShMem_module->defaultInstGate;
                    }
                    else {
                        /* Use the GateMP's default system gate */
                        gateHandle = GateMP_getDefaultRemote();
                        Assert_isTrue ((NULL != gateHandle),
                                        FrameQBufMgr_ShMem_A_nullHandle);
                    }
                }
                else {
                    /* Set flag  indicaring user has not provided  GateMP */
                    obj->creGate = TRUE;
                    /*Create gateMP for this instance */
                    GateMP_Params_init(&gateParams);
                    gateParams.localProtect = (GateMP_LocalProtect)
                                                          params->localProtect;
                    gateParams.remoteProtect = (GateMP_RemoteProtect)
                                                          params->remoteProtect;
                    gateParams.regionId = obj->regionId;
                    obj->gateMPAllocSize = GateMP_sharedMemReq(&gateParams);

                    obj->instGateSharedAddr =
                    gateParams.sharedAddr = Memory_alloc(
                                            SharedRegion_getHeap(obj->regionId),
                                            obj->gateMPAllocSize,
                                            0,
                                            eb);
                    Assert_isTrue ((NULL != gateParams.sharedAddr),
                                    FrameQBufMgr_ShMem_A_nullPointer);

                    gateHandle = GateMP_create (&gateParams);

                    Assert_isTrue ((NULL != gateHandle),
                                    FrameQBufMgr_ShMem_A_nullHandle);
                    /* Store the instance gate control address in  name server so
                     * that openers can open it.
                     */
                }
                obj->gate = gateHandle;
                nameserverEntry.instParams.shMemParams.
                                 instGateMPAddr =
                                          (Ptr)ti_sdo_ipc_GateMP_getSharedAddr(obj->gate);
            }
            /* Inserting nameserver entry params at the start of the shared
             * addr to allow open by  address.
             */
            pnameserverEntry = obj->pnameserverEntry;
            memcpy (pnameserverEntry,
                    &nameserverEntry,
                    sizeof (FrameQBufMgr_NameServerEntry));

            if (obj->ctrlStructCacheFlag == TRUE ) {
                Cache_wbInv ((Ptr)pnameserverEntry,
                             sizeof(FrameQBufMgr_NameServerEntry),
                             Cache_Type_ALL,
                             TRUE);
            }
            obj->minAlign = Memory_getMaxDefaultTypeAlign();

            if (SharedRegion_getCacheLineSize(obj->regionId) > obj->minAlign) {
                obj->minAlign = SharedRegion_getCacheLineSize(obj->regionId);
            }

            obj->ctrlInterfaceType  =
                           params->commonCreateParams.ctrlInterfaceType;
            obj->headerInterfaceType   =
                            params->headerInterfaceType;
            obj->bufInterfaceType   =
                            params->bufInterfaceType;

            obj->hdrBuf_regionId = params->hdrBuf_regionId;
            obj->frmBuf_regionId = params->frmBuf_regionId;
            /* Update cache flags based on shared region cache flags and
             * cpuAccessFlags for frame headers and frame buffers..
             */
            FrameQBufMgr_Shmem_updateCacheFlags(obj,
                                                params->cpuAccessFlags);

            /* Move addrOffset by cache aligned size of
             * FrameQ_NameServerEntry.
             */
            addrOffset += ROUND_UP(sizeof (FrameQBufMgr_NameServerEntry),
                                   obj->minAlign);

            obj->attrs = (FrameQBufMgr_ShMem_Attrs*) (virtAddr + addrOffset);

            pAttrs = obj->attrs;
            pAttrs->numNotifyEntries =  params->numNotifyEntries;

            /* Assuming FrameQBufMgr_ShMem_ControlStruct is
             * of cache aligned size
             */
            addrOffset += ROUND_UP(sizeof (FrameQBufMgr_ShMem_Attrs),
                                   obj->minAlign);
            obj->freeFramePool = Memory_alloc(NULL,
                                              (sizeof(UInt32) * params->numFreeFramePools),
                                              0,
                                              eb);
            Assert_isTrue ((NULL != obj->freeFramePool),
                            FrameQBufMgr_ShMem_A_nullPointer);

            for (i = 0; i < params->numFreeFramePools; i++) {
                obj->freeFramePool[i] = (FrameQBufMgr_ShMem_FreeFramePoolObj*)
                                                        (virtAddr + addrOffset);
                addrOffset += ROUND_UP(sizeof (FrameQBufMgr_ShMem_FreeFramePoolObj),
                                       obj->minAlign);
            }

            ClientNotifyMgr_Params_init(&clientMgrParams);
            clientMgrParams.numNotifyEntries = params->numNotifyEntries;
            clientMgrParams.numSubNotifyEntries = params->numFreeFramePools;
           /*
            * Pass the shared memory address to create the client
            * manager instance. This address stars right after the
            * size of (FrameQBufMgr_ShMem_ControlStruct) from the
            * params->sharedAddr.
            */
            clientMgrParams.sharedAddr = (Ptr)( virtAddr + addrOffset );
            clientMgrParams.sharedAddrSize =
                    ClientNotifyMgr_sharedMemReq (
                           (ClientNotifyMgr_Params*)&clientMgrParams );

            addrOffset += ROUND_UP(clientMgrParams.sharedAddrSize,
                                   obj->minAlign);
            if (params->gate != NULL) {
                clientMgrParams.gate = (Ptr) params->gate;
                /* Make it null so that finalization won't delete it.
                 */
                nameserverEntry.instParams.shMemParams.cliGateMPAddr =
                (Ptr) ti_sdo_ipc_GateMP_getSharedAddr((ti_sdo_ipc_GateMP_Handle)clientMgrParams.gate);

                obj->clientNotifyMgrGate = params->gate;
                obj->creCliGate = FALSE;
            }
            else {
                if (FrameQBufMgr_ShMem_usedefaultgate == TRUE ) {
                    clientMgrParams.gate = (Ptr) gateHandle;
                    /* Make it null so that finalization won't delete it.
                     */
                    nameserverEntry.instParams.shMemParams.cliGateMPAddr =
                    (Ptr) ti_sdo_ipc_GateMP_getSharedAddr((ti_sdo_ipc_GateMP_Handle)clientMgrParams.gate);
                    obj->clientNotifyMgrGate = NULL;
                    obj->creCliGate = FALSE;
                }
                else {
                    /*Create gateMP for the clientNotifyMgr */
                    obj->creCliGate = TRUE;
                    GateMP_Params_init(&gateParams);
                    gateParams.localProtect = (GateMP_LocalProtect)params->localProtect;
                    gateParams.remoteProtect = (GateMP_RemoteProtect)
                                                              params->remoteProtect;
                    gateParams.regionId = obj->regionId;
                    gateParams.sharedAddr  = (Ptr)(virtAddr + addrOffset );
                    gateMPsharedAddrSize = GateMP_sharedMemReq(&gateParams);

                    addrOffset += ROUND_UP(gateMPsharedAddrSize,
                                           obj->minAlign);
                    gateHandle = GateMP_create (&gateParams);

                    Assert_isTrue ((NULL != gateHandle),
                                    FrameQBufMgr_ShMem_A_nullHandle);
                    if (gateHandle == NULL) {
                        status = FrameQBufMgr_E_CREATE_GATEMP;
                        Error_raise(eb,
                                    FrameQBufMgr_ShMem_E_GateMPCreate,
                                    NULL,
                                    0);
                    }
                    else {
                        clientMgrParams.gate = (Ptr) gateHandle;
                        nameserverEntry.instParams.shMemParams.
                           cliGateMPAddr = (Ptr)SharedRegion_invalidSRPtr();
                        /* Make it null so that finalization won't delete it.
                         */
                        obj->clientNotifyMgrGate = gateHandle;
                    }
                }
            }

            if (status >= 0) {
                /* ClientNotifyMgr instance creation and  notification
                 * related stuff.
                 */
                clientMgrParams.numNotifyEntries = pAttrs->numNotifyEntries;
                /* Set the sub notify entries equqvent to number frame pools
                 */
                clientMgrParams.numSubNotifyEntries =
                                                  params->numFreeFramePools;
                clientMgrParams.name =
                                    (String)params->commonCreateParams.name;
                /* Event no to be used for this instance */
                clientMgrParams.eventNo =
                                 FrameQBufMgr_ShMem_NOTIFY_RESERVED_EVENTNO;
                /* This is actually create call */
                clientMgrParams.openFlag = FALSE;
                if (obj->ctrlStructCacheFlag == TRUE) {
                    clientMgrParams.cacheFlags =
                                               ClientNotifyMgr_CONTROL_CACHEUSE;
                }
                cliMgrHandle = ClientNotifyMgr_create (&clientMgrParams);
                Assert_isTrue ((NULL != cliMgrHandle),
                                FrameQBufMgr_ShMem_A_nullHandle);

                if (NULL == cliMgrHandle) {
                    status = FrameQBufMgr_E_FAIL_CLIENTN0TIFYMGR_CREATE;
                    Error_raise(eb,
                                FrameQBufMgr_ShMem_E_ClientNotifyMgrCreate,
                                NULL,
                                0);
                }
               else  {
                    /*********************************************************
                     *Memory manager allocaters handling
                     *********************************************************
                     */
                    tmpStatus = _FrameQBufMgr_ShMem_openHdrBufMemMgr(
                                                         obj,
                                                         pnameserverEntry,
                                                         eb);
                    if (tmpStatus < 0) {
                        status = tmpStatus;
                        /* Pass the generated error up  */
                    }
                    else {
                        tmpStatus = _FrameQBufMgr_ShMem_openFrmBufMemMgr(
                                                             obj,
                                                             pnameserverEntry,
                                                             eb);
                        if (tmpStatus < 0) {
                            status = tmpStatus;
                            /* Pass the generated error up  */
                        }
                        else {
                            obj->clientNotifyMgrHandle = (Ptr)cliMgrHandle;
                            if (params->sharedAddr) {
                                obj->objType   =
                                              FrameQBufMgr_ShMem_DYNAMIC_CREATE;
                            }
                            else {
                                /* set the object type indicating it is created
                                 * using Shared Region id internally.
                                 */
                                obj->objType  =
                                   FrameQBufMgr_ShMem_DYNAMIC_CREATE_USEDREGION;
                            }

                            obj->isRegistered   = FALSE;
                            obj->notifyId       = (UInt32)-1;
                            obj->instId = pAttrs->entryId = instId;
                            strcpy((String)pAttrs->name ,
                                   (String)params->commonCreateParams.name);
                            pAttrs->name[31] = '\0';

                            strcpy((Ptr)obj->name ,
                                   (const Char*)pAttrs->name);

                            pAttrs->version = FrameQBufMgr_ShMem_VERSION;
                            pAttrs->numFreeFramePools =
                                                      params->numFreeFramePools;
                            pAttrs->localProtect = params->localProtect;
                            pAttrs->remoteProtect = params->remoteProtect;
                            obj->refCount = 1;

                            /* Populate Bufpools if buffer chunks are not
                             * provided This API allocates buffers
                             */
                            status = FrameQBufMgr_ShMem_postInit (obj,
                                                                  params,
                                                                  eb);
                            if (status < 0) {
                                Error_raise(eb,
                                            FrameQBufMgr_ShMem_E_postInitFail,
                                            NULL,
                                            0);
                            }
                            else {
                               /* Add name to the FrameQBufMgr name server
                                * instance.
                                * The Name server instance is created in  module
                                * initialization  phase.
                                */
                                key = GateMP_enter ((GateMP_Handle)obj->gate);
                                nameLen =  strlen(
                                       (String)params->commonCreateParams.name);
                                if (   (   NULL
                                       != FrameQBufMgr_ShMem_module->nameServer)
                                    && (nameLen > 0)) {
                                    /* Populate the nameserver entry  with the
                                     * params.
                                     */
                                    obj->nsKey = NameServer_add ((NameServer_Handle)
                                        FrameQBufMgr_ShMem_module->nameServer,
                                        (String)params->commonCreateParams.name,
                                        &nameserverEntry,
                                         sizeof(FrameQBufMgr_NameServerEntry));
                                    if (obj->nsKey == NULL) {
                                        status = FrameQBufMgr_E_NAMESERVERADD;
                                        Assert_isTrue ((NULL != obj->nsKey),
                                                       FrameQBufMgr_ShMem_A_nullPointer);
                                    }
                                }
                                GateMP_leave ((GateMP_Handle)obj->gate, key);
                            }
                        }
                    }
                }
            }
            if (status < 0) {
                status = FrameQBufMgr_freeInstNo (instNo) ;
            }
        }
        else {
            Error_raise(eb,
                        FrameQBufMgr_ShMem_E_maxInstances,
                        NULL,
                        0);
        }
    }
    return (status);
}

/*
 *  ======== _FrameQBufMgr_ShMem_open ========
 *  Internal API to open the instance.Called internally by
 *  the FrameQBufMgr_ShMem_instance_init() call.
 */
static inline
Int32
_FrameQBufMgr_ShMem_open(FrameQBufMgr_ShMem_Object       * obj,
                         const ti_syslink_ipc_rtos_shMem_FrameQBufMgr_ShMem_Params * params,
                         Error_Block *eb)
{
    Int32                         status                = FrameQBufMgr_S_SUCCESS;
    volatile FrameQBufMgr_ShMem_Attrs         * pAttrs  = NULL;
    UInt32                        addrOffset     = 0;
    UInt32                        virtAddr       = 0;
    Int32                         tmpStatus;
    ClientNotifyMgr_Params        clientMgrParams;
    ClientNotifyMgr_Handle        cliMgrHandle;
    UInt32                        regionId;
    GateMP_Params                 gateParams;
    GateMP_Handle                 gateHandle;
    UInt8                         nameLen;
    FrameQBufMgr_NameServerEntry  nameserverEntry;
    FrameQBufMgr_NameServerEntry *pnameserverEntry;
    Ptr                           localAddr;
    UInt32                        len;
    UInt32                        key;
    UInt32                        i;

    Assert_isTrue ((   (NULL != (String)params->commonCreateParams.name)
                     ||(NULL != params->sharedAddr)),
                    FrameQBufMgr_ShMem_A_invalidParams);

    /* Open the already created instance */
    if ((String)params->commonCreateParams.name != NULL) {
        nameLen = strlen((String)params->commonCreateParams.name);

        Assert_isTrue ((FrameQBufMgr_ShMem_MAX_NAMELEN >= (nameLen + 1)),
                        FrameQBufMgr_ShMem_A_invalidParams);

        Assert_isTrue ((NULL != FrameQBufMgr_ShMem_module->nameServer),
                        FrameQBufMgr_ShMem_A_invalidParams);

        len = sizeof (FrameQBufMgr_NameServerEntry);
        status = NameServer_get ((NameServer_Handle)FrameQBufMgr_ShMem_module->nameServer,
                                 (String)params->commonCreateParams.name,
                                 &nameserverEntry,
                                 &len,
                                 NULL);
        if (status == NameServer_E_NOTFOUND) {
            /* Name not found. */
            /* Failed to get the information from the name server.
             * It means it is not created.
             */
            status = FrameQBufMgr_E_NOTFOUND;
            Error_raise(eb,
                        FrameQBufMgr_ShMem_E_FrameQBufMgrNotFound,
                        NULL,
                        0);
        }
        else if (status == NameServer_E_FAIL) {
            /* Error happened in NameServer. Pass the error up. */
            status = FrameQBufMgr_E_FAIL;
            Error_raise(eb,
                        FrameQBufMgr_E_InvalidParams,
                        NULL,
                        0);
        }
        else {
            virtAddr = (UInt32) SharedRegion_getPtr (
                                      (SharedRegion_SRPtr)
                 nameserverEntry.instParams.shMemParams.sharedAddr);
            pnameserverEntry = (FrameQBufMgr_NameServerEntry*)virtAddr;
            /* Assert that sharedAddr is cache aligned */
            regionId = SharedRegion_getId(pnameserverEntry);

            obj->regionId = regionId;
            /* Assert that the buffer is in a valid shared region */
            Assert_isTrue(regionId != SharedRegion_INVALIDREGIONID,
                          Ipc_A_addrNotInSharedRegion);

            Assert_isTrue(((UInt32)pnameserverEntry %
                          SharedRegion_getCacheLineSize(regionId) == 0),
                          Ipc_A_addrNotCacheAligned);
        }
    }
    else if (params->sharedAddr != NULL) {
        /* Unable to get information from name server. But user has
         * provided  virtual shared address. So proceeding further.
         */
        virtAddr    = (UInt32) params->sharedAddr;
        pnameserverEntry = (FrameQBufMgr_NameServerEntry*)virtAddr;
        /* Invalidate the  FrameQBufMgr entry info located at the start of
         * the sharedAddr.
         */
        /* Assert that sharedAddr is cache aligned */
        regionId = SharedRegion_getId(pnameserverEntry);
        obj->regionId = regionId;
        /* Assert that the buffer is in a valid shared region */
        Assert_isTrue(regionId != SharedRegion_INVALIDREGIONID,
                      Ipc_A_addrNotInSharedRegion);

        Assert_isTrue(((UInt32)pnameserverEntry %
                      SharedRegion_getCacheLineSize(regionId) == 0),
                      Ipc_A_addrNotCacheAligned);

        if (SharedRegion_isCacheEnabled(regionId)) {
            Cache_inv((Ptr)pnameserverEntry,
                      ROUND_UP (
                              sizeof (FrameQBufMgr_NameServerEntry),
                              SharedRegion_getCacheLineSize(obj->regionId)),
                      Cache_Type_ALL,
                      TRUE);
        }
        memcpy(&nameserverEntry,
               pnameserverEntry,
               sizeof(FrameQBufMgr_NameServerEntry));
        status      = FrameQBufMgr_S_SUCCESS;
    }
    else {
        /*params->sharedAddr is null and also virtaddr got from the
         * name server is  also not valid. So rejecting the  open
         *call.
         */
        status = FrameQBufMgr_E_FAIL;
        Error_raise(eb,
                    FrameQBufMgr_ShMem_E_FrameQBufMgrNotCreated,
                    NULL,
                    0);
    }

    if (status >= 0) {
        Assert_isTrue (   (nameserverEntry.instParams.shMemParams.
                           instGateMPAddr != (Ptr)SharedRegion_invalidSRPtr()),
                       FrameQBufMgr_ShMem_A_invalidParams);

        if (   nameserverEntry.instParams.shMemParams.instGateMPAddr
            != (Ptr)SharedRegion_invalidSRPtr()) {
            /* Get the local address of the SRPtr */
            localAddr = SharedRegion_getPtr((SharedRegion_SRPtr)
                nameserverEntry.instParams.shMemParams.instGateMPAddr);

            /* Open gateMP using the shared Addr of the gateMP created
             * by the  FrameQ creator
             */
            tmpStatus = GateMP_openByAddr(localAddr, &gateHandle);
            if (tmpStatus != GateMP_S_SUCCESS) {
                obj->creGate = FALSE;
                status = FrameQBufMgr_E_OPEN_GATEMP;
                Error_raise(eb,
                    FrameQBufMgr_ShMem_E_GateMPOpen,
                    NULL,
                    0);
            }
            else {
                obj->gate    = gateHandle;
                obj->creGate = TRUE;
            }
        }
    }

    if (status >= 0) {
        /* Set the Id to invalid */
        obj->instId = (UInt32)-1;
        obj->minAlign = Memory_getMaxDefaultTypeAlign();

        if (SharedRegion_getCacheLineSize(obj->regionId) > obj->minAlign) {
            obj->minAlign = SharedRegion_getCacheLineSize(obj->regionId);
        }

        obj->ctrlInterfaceType     =
                       nameserverEntry.ctrlInterfaceType;
        obj->headerInterfaceType   =
                        nameserverEntry.headerInterfaceType;
        obj->bufInterfaceType      =
                        nameserverEntry.bufInterfaceType;

        obj->hdrBuf_regionId =
                        nameserverEntry.instParams.shMemParams.hdrBuf_regionId;
        obj->frmBuf_regionId =
                        nameserverEntry.instParams.shMemParams.frmBuf_regionId;
        /* Update the control cache flag */
        obj->ctrlStructCacheFlag =
                    SharedRegion_isCacheEnabled(obj->regionId);

        /* Update cache flags based on shared region cache flags and
         * cpuAccessFlags.
         */
        FrameQBufMgr_Shmem_updateCacheFlags(obj,
                                            params->cpuAccessFlags);

        /* Move offset by size of nameserverEntry  cache aligned size.
         * No need to check interface type as it is checked by front end
         * module(FrameQBufMgr).
         */
        obj->pnameserverEntry = (FrameQBufMgr_NameServerEntry*)(virtAddr);
        addrOffset += ROUND_UP(sizeof (FrameQBufMgr_NameServerEntry),
                              obj->minAlign);

        obj->attrs = (FrameQBufMgr_ShMem_Attrs*) (virtAddr + addrOffset);

        addrOffset += ROUND_UP(sizeof (FrameQBufMgr_ShMem_Attrs),
                              obj->minAlign);

        /* Enter the gate */
        key = GateMP_enter ((GateMP_Handle)obj->gate);

         /*Invalidate the  virtAddr contents */
        pAttrs = obj->attrs;
        if (obj->ctrlStructCacheFlag == TRUE) {
            Cache_inv((Ptr)obj->pnameserverEntry,
                      sizeof (FrameQBufMgr_NameServerEntry),
                      Cache_Type_ALL,
                      TRUE);
            Cache_inv((Ptr)pAttrs,
                      sizeof (FrameQBufMgr_ShMem_Attrs),
                      Cache_Type_ALL,
                      TRUE);
        }
        /* Make sure the instance has been created */
        if (pAttrs->status != FrameQBufMgr_ShMem_CREATED) {
            Error_raise(eb,
                        FrameQBufMgr_ShMem_E_FrameQBufMgrNotCreated,
                        NULL,
                        0);
            status = FrameQBufMgr_E_NOTFOUND;
        }
        else {
            obj->freeFramePool = Memory_alloc(NULL,
                                              (sizeof(UInt32) * pAttrs->numFreeFramePools),
                                              0,
                                              eb);
            if (obj->freeFramePool != NULL) {
                for (i = 0; i < pAttrs->numFreeFramePools; i++) {
                    obj->freeFramePool[i] = (FrameQBufMgr_ShMem_FreeFramePoolObj*)
                                                            (virtAddr + addrOffset);
                    addrOffset += ROUND_UP(sizeof (FrameQBufMgr_ShMem_FreeFramePoolObj),
                                           obj->minAlign);
                }

                ClientNotifyMgr_Params_init(&clientMgrParams);
                clientMgrParams.numNotifyEntries = pAttrs->numNotifyEntries;
                clientMgrParams.numSubNotifyEntries = pAttrs->numFreeFramePools;
                clientMgrParams.sharedAddr = (Ptr)
                                          (  virtAddr
                                           + addrOffset );
                clientMgrParams.sharedAddrSize =
                        ClientNotifyMgr_sharedMemReq (
                               (ClientNotifyMgr_Params*)&clientMgrParams );
                addrOffset += ROUND_UP(clientMgrParams.sharedAddrSize,obj->minAlign);
                /* cliGateMPAddr if it valid means user has passed the
                 * gate to create call or  default gate is used in create call.
                 */
                if (   nameserverEntry.instParams.shMemParams.cliGateMPAddr
                    != (Ptr)SharedRegion_invalidSRPtr()) {
                    /* Get the local address of the SRPtr */
                    localAddr = SharedRegion_getPtr((SharedRegion_SRPtr)
                         nameserverEntry.instParams.shMemParams.cliGateMPAddr);
                    Assert_isTrue ((localAddr != NULL),
                                FrameQBufMgr_ShMem_A_nullPointer);

                    tmpStatus = GateMP_openByAddr(localAddr, &gateHandle);
                    if (tmpStatus != GateMP_S_SUCCESS) {
                        status = FrameQ_E_OPEN_GATEMP;
                        obj->creCliGate = FALSE;
                        /* Make it null so that finalization won't delete it.
                         */
                        obj->clientNotifyMgrGate = NULL;
                    }
                    else {
                        clientMgrParams.gate = (Ptr)gateHandle;
                        obj->clientNotifyMgrGate  = gateHandle;
                        obj->creCliGate = TRUE;
                    }
                }
                else {
                    /* Open the gate created for the ClientNotifyMgr instance*/
                    obj->creCliGate = TRUE;
                    GateMP_Params_init(&gateParams);

                    gateParams.localProtect  =
                                  (GateMP_LocalProtect)pAttrs->localProtect;
                    gateParams.remoteProtect =
                                    (GateMP_RemoteProtect)pAttrs->remoteProtect;
                    gateParams.regionId = obj->regionId;
                    gateParams.sharedAddr  = (Ptr)(virtAddr + addrOffset );

                    addrOffset += ROUND_UP(GateMP_sharedMemReq(&gateParams),
                                           obj->minAlign);
                    tmpStatus = GateMP_openByAddr(gateParams.sharedAddr,
                                              &gateHandle);
                    if (tmpStatus != GateMP_S_SUCCESS) {
                        /* Error happened in GateMP*/
                        obj->clientNotifyMgrGate = NULL;
                        status = FrameQBufMgr_E_OPEN_GATEMP;
                        Error_raise(eb,
                                    FrameQBufMgr_ShMem_E_GateMPOpen,
                                    NULL,
                                    0);
                    }
                    else {
                        clientMgrParams.gate = (Ptr)gateHandle;
                        obj->clientNotifyMgrGate  = gateHandle;
                    }
                }
                if (status >= 0) {
                    /* Open  the already created clientNotifyMgr  instance  using
                     * Shared addr
                     */
                    if (obj->ctrlStructCacheFlag == TRUE) {
                        clientMgrParams.cacheFlags =
                                            ClientNotifyMgr_CONTROL_CACHEUSE;
                    }

                    /*Open the client notifyMgr instace*/
                    clientMgrParams.openFlag = TRUE;
                    cliMgrHandle = ClientNotifyMgr_create (&clientMgrParams);
                    if (cliMgrHandle == NULL) {
                        status = FrameQBufMgr_E_CLIENTN0TIFYMGR_OPEN;
                        Error_raise(eb,
                                    FrameQBufMgr_ShMem_E_ClientNotifyMgrOpen,
                                    NULL,
                                    0);
                    }
                    else {
                        obj->clientNotifyMgrHandle = (Ptr)cliMgrHandle;
                        obj->objType               =
                                                    FrameQBufMgr_ShMem_DYNAMIC_OPEN;
                        obj->isRegistered          = FALSE;
                        obj->notifyId              = (UInt32)-1;
                        obj->refCount              = 1;
                        obj->instId                = pAttrs->entryId;
                        strcpy((String)obj->name ,(const Char *)pAttrs->name);
                        /*********************************************************
                         *Memory manager allocaters handling
                         *********************************************************
                         */
                        status = _FrameQBufMgr_ShMem_openHdrBufMemMgr(
                                                             obj,
                                                             pnameserverEntry,
                                                             eb);
                        if (status >= 0) {
                            status = _FrameQBufMgr_ShMem_openFrmBufMemMgr(
                                                                 obj,
                                                                 pnameserverEntry,
                                                                 eb);
                        }
                    }
                }
            }
            else {
                status  = FrameQBufMgr_E_FAIL;
                Error_raise(eb,
                            FrameQBufMgr_ShMem_E_MemoryAlloc,
                            NULL,
                            0);
            }
        }
        /* Leave the gate */
        GateMP_leave ((GateMP_Handle)obj->gate, key);
    }

    return (status);
}

/*
 *  ======== _FrameQBufMgr_ShMem_openHdrBufMemMgr ========
 *  Internal API to open the instance.Called internally.
 */
static inline
Int32
_FrameQBufMgr_ShMem_openHdrBufMemMgr(
                                FrameQBufMgr_ShMem_Object *obj,
                                FrameQBufMgr_NameServerEntry *pnameserverEntry,
                                Error_Block *eb)
{
    Int32                         status         = FrameQBufMgr_S_SUCCESS;
    SyslinkMemMgr_Params          hdrBufMemMgrParams;

    /**********************************************************
     * Memory manager creation for Header bufs
     **********************************************************
     */
    switch (pnameserverEntry->headerInterfaceType) {
        case FrameQBufMgr_HDRINTERFACE_SHAREDMEM:
        {
            /* Open HeapMemMP that is needed for HeaderBuffers
             * Open only if name is provided.
             */
            if (status >= 0) {
                /* Initialize params for the shared  Memory manager
                 */
                SyslinkMemMgr_Params_init (&hdrBufMemMgrParams);
                hdrBufMemMgrParams.memMgrType = SyslinkMemMgr_TYPE_SHAREDMEM;

                /* Pass null to the MemMgr as we are explicitly
                 * providing memory for headerbufs.
                 */
                hdrBufMemMgrParams.heapHandle = SharedRegion_getHeap((UInt16)
                     pnameserverEntry->instParams.shMemParams.hdrBuf_regionId);
                obj->heapMemMPHandle_HdrBufs = hdrBufMemMgrParams.heapHandle;

                obj->memMgrAllocater_hdrBufs = SyslinkMemMgr_create(&hdrBufMemMgrParams,eb);
                obj->memMgrType_HdrBufs = SyslinkMemMgr_TYPE_SHAREDMEM;
                Assert_isTrue ((NULL != obj->memMgrAllocater_hdrBufs),
                               FrameQBufMgr_ShMem_A_nullHandle);
            }
        }
        break;

        default:
            Assert_isTrue (0, FrameQBufMgr_ShMem_A_invalid_MemMgrType);

            Error_raise(eb,
                        FrameQBufMgr_ShMem_E_InvalidMemMgrType,
                        NULL,
                        0);
            status = FrameQBufMgr_E_INVALID_HDRINTERFACETYPE;
    }

    return (status);
}

/*
 *  ======== _FrameQBufMgr_ShMem_openFrmBufMemMgr ========
 *  Internal API to open the instance.Called internally.
 */
static inline
Int32
_FrameQBufMgr_ShMem_openFrmBufMemMgr(
                                 FrameQBufMgr_ShMem_Object *obj,
                                 FrameQBufMgr_NameServerEntry *pnameserverEntry,
                                 Error_Block *eb)
{
    Int32                         status         = FrameQBufMgr_S_SUCCESS;
    SyslinkMemMgr_Params          frmBufMemMgrParams;
    /**********************************************************
     * Memory manager creation for frame buffers.
     **********************************************************
     */

    /* Initialize params for the shared  Memory manager
     */
    SyslinkMemMgr_Params_init (&frmBufMemMgrParams);

    switch (pnameserverEntry->bufInterfaceType) {
        case FrameQBufMgr_BUFINTERFACE_SHAREDMEM:
        {
            /* Open HeapMemMP that is needed for HeaderBuffers
             * Open only if name is provided.
             */


                frmBufMemMgrParams.heapHandle = SharedRegion_getHeap((UInt16)
                     pnameserverEntry->instParams.shMemParams.frmBuf_regionId);
                obj->heapMemMPHandle_FrmBufs = frmBufMemMgrParams.heapHandle;

                frmBufMemMgrParams.memMgrType = SyslinkMemMgr_TYPE_SHAREDMEM;

               /* make it based on bufer interface type*/
                obj->memMgrAllocater_frmBufs = SyslinkMemMgr_create(&frmBufMemMgrParams,eb);
                obj->memMgrType_FrmBufs = SyslinkMemMgr_TYPE_SHAREDMEM;
                Assert_isTrue ((NULL != obj->memMgrAllocater_frmBufs),
                               FrameQBufMgr_ShMem_A_nullHandle);
        }
        break;
        case FrameQBufMgr_BUFINTERFACE_TILERMEM:
        {
            /* Open HeapMemMP that is needed for HeaderBuffers
             * Open only if name is provided.
             */
            frmBufMemMgrParams.memMgrType = SyslinkMemMgr_TYPE_TILERMEM;

            /* make it based on bufer interface type*/
            obj->memMgrAllocater_frmBufs = SyslinkMemMgr_create(&frmBufMemMgrParams, eb);
            obj->memMgrType_FrmBufs = SyslinkMemMgr_TYPE_TILERMEM;
            Assert_isTrue ((NULL != obj->memMgrAllocater_frmBufs),
                           FrameQBufMgr_ShMem_A_nullHandle);

        }
        break;
        default:
            Assert_isTrue (0,
                           FrameQBufMgr_ShMem_A_invalid_MemMgrType);
            Error_raise(eb,
                        FrameQBufMgr_ShMem_E_InvalidMemMgrType,
                        NULL,
                        0);
            status = FrameQBufMgr_E_INVALID_BUFINTERFACETYPE;
    }

    return (status);
}

/*
 *  ======== FrameQBufMgr_Shmem_updateCacheFlags ========
 *  Internal API to populate the cache flags
 */
Void FrameQBufMgr_Shmem_updateCacheFlags(FrameQBufMgr_ShMem_Object *obj,
                                         UInt32 cpuAccessFlags)
{
    UInt32  i = 0;
    UInt32  bufCacheFlag;

    /* Derieving Header buffer cache flag from the shared region id */
    obj->frmHdrBufCacheFlag = SharedRegion_isCacheEnabled(obj->hdrBuf_regionId);

    /* Derieving frame buffer cache flags from the shared region id and
     * cpuAccessFlags.
     */
    if (obj->bufInterfaceType == FrameQBufMgr_BUFINTERFACE_SHAREDMEM) {
        bufCacheFlag = SharedRegion_isCacheEnabled(obj->frmBuf_regionId);
    }
    else if (obj->bufInterfaceType == FrameQBufMgr_BUFINTERFACE_TILERMEM) {
        /* TBD: Currently Tiler buffers are from non cached memory */
        bufCacheFlag = FALSE;
    }
    else {
    }

    for (i = 0; i < FrameQBufMgr_ShMem_MAX_FRAMEBUFS; i++) {
        /* Check and update if user has provided cache  flags for Frame
         * buffers.
         */
        obj->frmBufCacheFlag[i] = bufCacheFlag;

        /* Check and update if user has provided cpu access flags for Frame
         * buffers.
         */
        if (   cpuAccessFlags
             & (1 << (FrameQBufMgr_ShMem_FRAMEBUF_CPUACCESSFLAGS_BITOFFSET +i))) {
            obj->bufCpuAccessFlag[i] = TRUE;
        }
    }
}

/*
 *  ======== FrameQBufMgr_postInit ========
 *  Internal API to populate the frame pools.
 */
Int32
FrameQBufMgr_ShMem_postInit(FrameQBufMgr_ShMem_Object       * obj,
                            const ti_syslink_ipc_rtos_shMem_FrameQBufMgr_ShMem_Params * params,
                            Error_Block *eb)
{
    Int32                         status     = FrameQBufMgr_S_SUCCESS;
    UInt32                        addrOffset = 0;
    FrameQBufMgr_ShMem_FrameBufInfo              *frameBufInfo  = NULL;
    volatile FrameQBufMgr_ShMem_Attrs            *pAttrs;
    volatile FrameQBufMgr_ShMem_FreeFramePoolObj *freeFramePoolobj;
    UInt32                        i;
    UInt32                        j;
    UInt32                        k;
    UInt32                        memSzForCliMgr;
    ClientNotifyMgr_Params        clientMgrParams;
    FrameQBufMgr_ShMem_Frame      frame;
    UInt32                        lhdrChunkStartAddr;
    UInt32                        lfrmChunkStartAddr;
    UInt32                        lhdrChunkSize;
    UInt32                        lfrmChunkSize;
    UInt32                        numframes;
    UInt32                        numHdrs;
    UInt32                        hdrSize;
    UInt32                        numBufsInFrm;
    UInt32                        frmbufOffset;
    UInt32                        tframbufSize;
    UInt32                        virtAddr;
    UInt32                        hdrBufMemReq;
    UInt32                        frmBufMemReq;
    SyslinkMemMgr_allocParams     allocParams;
    SharedRegion_Entry            srEntry;

    pAttrs = obj->attrs;

    if (params->numFreeFramePools) {

        /* Start of the instance shared addr */
        virtAddr = (UInt32)obj->pnameserverEntry;

        /* ---------------------------------------------------------
         * Update the control structure for number of frame pools
         * ---------------------------------------------------------
         */
        pAttrs->numFreeFramePools = params->numFreeFramePools;

        /* To create the  list objects */
        addrOffset += ROUND_UP(sizeof (FrameQBufMgr_NameServerEntry),
                              obj->minAlign);


        addrOffset += ROUND_UP(sizeof (FrameQBufMgr_ShMem_Attrs),
                              obj->minAlign);

        addrOffset += ROUND_UP(sizeof (FrameQBufMgr_ShMem_FreeFramePoolObj),
                              obj->minAlign) * params->numFreeFramePools;

        addrOffset += ROUND_UP(obj->cliGateMPShAddrSize,
                              obj->minAlign);

        ClientNotifyMgr_Params_init(&clientMgrParams);
        clientMgrParams.regionId = SharedRegion_getId((Ptr)pAttrs);
        clientMgrParams.sharedAddr = (UInt32*)((UInt32)virtAddr
                                                  + addrOffset );
        clientMgrParams.numNotifyEntries = pAttrs->numNotifyEntries;
        clientMgrParams.numSubNotifyEntries = pAttrs->numFreeFramePools;

        memSzForCliMgr =  ClientNotifyMgr_sharedMemReq (
                               (ClientNotifyMgr_Params*)&clientMgrParams );

        addrOffset += ROUND_UP(memSzForCliMgr, obj->minAlign);

        ti_syslink_ipc_rtos_shMem_FrameQBufMgr_ShMem_sharedMemReq__F
                 (params, &hdrBufMemReq, &frmBufMemReq);

        if (params->sharedAddrHdrBuf == NULL) {
            /* Allocate memory from the hdrBuf_regionId */
            obj->hdrBuf_regionId = params->hdrBuf_regionId;
            obj->hdrBufAllocSize = hdrBufMemReq;
            obj->hdrBufPtr  = Memory_alloc( SharedRegion_getHeap(obj->hdrBuf_regionId),
                                            obj->hdrBufAllocSize,
                                            0,
                                            eb);
            if (obj->hdrBufPtr != NULL) {
                lhdrChunkStartAddr = (UInt32)obj->hdrBufPtr;
                lhdrChunkSize = obj->hdrBufAllocSize;
            }
            else {
                status  = FrameQBufMgr_E_FAIL;
                Error_raise(eb,
                            FrameQBufMgr_ShMem_E_MemoryAlloc,
                            NULL,
                            0);
            }
        }
        else {
            obj->hdrBufPtr = params->sharedAddrHdrBuf;
            obj->hdrBufAllocSize = 0;
            lhdrChunkStartAddr = (UInt32)params->sharedAddrHdrBuf;
            lhdrChunkSize = params->sharedAddrHdrBufSize;
        }

        if (params->sharedAddrFrmBuf == NULL) {
            /* Allocate memory from the hdrBuf_regionId */
            obj->frmBuf_regionId = params->frmBuf_regionId;
            obj->frmBufAllocSize = frmBufMemReq;

            if(obj->bufInterfaceType ==
                    FrameQBufMgr_ShMem_BUFINTERFACE_SHAREDMEM){

                SharedRegion_getEntry(obj->frmBuf_regionId, &srEntry);
                /* no need to check SharedRegion_S_SUCCESS, asserted before */

                if (srEntry.createHeap == TRUE) {

                    obj->frmBufPtr = Memory_alloc(
                            SharedRegion_getHeap(obj->frmBuf_regionId),
                                                obj->frmBufAllocSize,
                                                0,
                                                eb);
                    if (obj->frmBufPtr == NULL) {
                        status  = FrameQBufMgr_E_FAIL;
                        Error_raise(eb,
                                FrameQBufMgr_ShMem_E_MemoryAlloc,
                                NULL,
                                0);
                    }
                    lfrmChunkSize = obj->frmBufAllocSize;

                } else {
                    /* This is the case when frame-buffer memory is not mapped
                     * to kernel virtual memory space on the HOST OS.
                     *
                     * On the RTOS side simply use the Shared Region Base
                     * address as the exclusive owner of that SR, and assign
                     * the whole region for frame buffer memory.
                     */
                    obj->frmBufPtr = (Ptr) srEntry.base;
                    obj->frmBufAllocSize = 0;
                    lfrmChunkSize =  srEntry.len;
                }
            }
            else{
                /* Will be assigned at MemMgr_alloc for tiled buffers*/
                obj->frmBufPtr  = NULL;
            }
            lfrmChunkStartAddr = (UInt32)obj->frmBufPtr;
        }
        else {
            obj->frmBufPtr = params->sharedAddrFrmBuf;
            obj->frmBufAllocSize = 0;
            lfrmChunkStartAddr = (UInt32)params->sharedAddrFrmBuf;
            lfrmChunkSize = params->sharedAddrFrmBufSize;
        }

        for ( i = 0;
              (i < params->numFreeFramePools) && (status >= 0);
              i++) {
            numframes = params->numFreeFramesInPool[i];
            numHdrs   = params->numFreeHeaderBufsInPool[i];
            hdrSize   = params->frameHeaderBufSize[i];
            numBufsInFrm  = params->numFrameBufsInFrame[i];
            /* Checking if hdr chunk is sufficient */
            if (  ((numframes + numHdrs) * hdrSize) + lhdrChunkStartAddr
                <= ((UInt32)obj->hdrBufPtr + lhdrChunkSize)) {
                /*  Now populate frames and header buffers */
                /* Check if frm buf chunk is sufficient */
                tframbufSize = 0;
                if(obj->bufInterfaceType == FrameQBufMgr_ShMem_BUFINTERFACE_SHAREDMEM){
                /*TBD:Check this condition forTilerMem*/
                    for (j = 0; j < numBufsInFrm ; j++) {
                        tframbufSize += params->frameBufParams[i][j].size;
                    }

                    if (  ((numframes * tframbufSize)+ lfrmChunkStartAddr)
                        > ((UInt32)obj->frmBufPtr + lfrmChunkSize)) {
                        status = FrameQBufMgr_E_FAIL;
                        Error_raise(eb,
                                    FrameQBufMgr_ShMem_E_Fail,
                                    NULL,
                                    0);
                    }
                }
                if (status >= 0) {
                    freeFramePoolobj = obj->freeFramePool[i];
                    /* Frame header buffer size of this pool */
                    freeFramePoolobj->frmHdrBufSize = hdrSize;

                    /* Number of frame buffers in a frame
                     * belonging to this pool
                     */
                    freeFramePoolobj->numBufsInFrame = numBufsInFrm;

                    for (k = 0; k < numBufsInFrm; k++) {
                        freeFramePoolobj->frmBufSize[k] =
                                                   params->frameBufParams[i][k].size;
                    }

                    /* Number of Extra headers in this pool */
                    freeFramePoolobj->numFrmHdrBufs =  numHdrs;

                    freeFramePoolobj->freeFrmHdrBufs =
                                           freeFramePoolobj->numFrmHdrBufs;

                    /* Number of frames in this pool */
                    freeFramePoolobj->numFrms = numframes;

                    freeFramePoolobj->freeFrms =
                                                freeFramePoolobj->numFrms;

                    /* Allow  dynamic addition of  frames and frame buffers to
                     * this pool.
                     */
                    freeFramePoolobj->numDynamicHdrBufs  = 0;
                    freeFramePoolobj->freeDynamicHdrBufs = 0;

                    freeFramePoolobj->numDynamicFrames  = 0;
                    freeFramePoolobj->freeDynamicFrames = 0;

                    for (j = 0; j < FrameQBufMgr_ShMem_POOL_MAXFRAMES; j++) {
                        freeFramePoolobj->frmRefCnt[j] = 0;
                    }

                    for (j = 0; j < FrameQBufMgr_ShMem_POOL_DYNAMIC_MAXFRAMES; j++)
                    {
                        freeFramePoolobj->dynFrmRefCnt[j] = 0;
                    }

                    for (j = 0; j < FrameQBufMgr_ShMem_POOL_MAXHDRS; j++) {
                        /* To search for free hdr buf  for dup purposes.
                         */
                        freeFramePoolobj->hdrBufIsUsed[j] = FALSE;
                        /* To keep track of  hdr buf is  pointing to  which
                         * reference count.
                         */
                        freeFramePoolobj->hdrBuf_refCntIndex[j] = (UInt32)-1;
                        freeFramePoolobj->hdrBuf_refCntIndexType[j] = 0;

                    }

                    for (j = 0;
                         j < FrameQBufMgr_ShMem_POOL_DYNAMIC_MAXHDRS;
                         j++) {
                        /* To search for free hdr buf  for dup purposes.
                         */
                        freeFramePoolobj->dynHdrBufIsUsed[j] = FALSE;
                        /* To keep track of  hdr buf is  pointing to  which
                         * reference count.
                         */
                        freeFramePoolobj->dynHdrBuf_refCntIndex[j] = (UInt32)-1;
                        freeFramePoolobj->dynHdrBuf_refCntIndexType[j] = 0;
                        freeFramePoolobj->dynHdrBufs[j] = NULL;
                    }

                    /* Populate frames and headers */
                    /* Convert  address in to portable address */
                    freeFramePoolobj->frmStartAddrp = (UInt32)
                            SyslinkMemMgr_translate (obj->memMgrAllocater_hdrBufs,
                                              (Ptr) lhdrChunkStartAddr,
                                              SyslinkMemMgr_AddrType_Virtual,
                                              SyslinkMemMgr_AddrType_Portable);

                    freeFramePoolobj->frmEndAddrp = (UInt32)
                            SyslinkMemMgr_translate (obj->memMgrAllocater_hdrBufs,
                                              (Ptr)( lhdrChunkStartAddr +(numframes * hdrSize)),
                                              SyslinkMemMgr_AddrType_Virtual,
                                              SyslinkMemMgr_AddrType_Portable);

                    freeFramePoolobj->hdrBufStartAddrp = (UInt32)
                            SyslinkMemMgr_translate (obj->memMgrAllocater_hdrBufs,
                                              (Ptr)( lhdrChunkStartAddr +(numframes * hdrSize)),
                                              SyslinkMemMgr_AddrType_Virtual,
                                              SyslinkMemMgr_AddrType_Portable);

                    freeFramePoolobj->hdrBufEndAddrp = (UInt32)
                            SyslinkMemMgr_translate (obj->memMgrAllocater_hdrBufs,
                                              (Ptr)( lhdrChunkStartAddr +((numframes + numHdrs) * hdrSize)),
                                              SyslinkMemMgr_AddrType_Virtual,
                                              SyslinkMemMgr_AddrType_Portable);
                    if(obj->bufInterfaceType
                            == FrameQBufMgr_ShMem_BUFINTERFACE_SHAREDMEM){

                        /*
                         * On the RTOS this translation will actually
                         * work whether allocation was from SR Heap
                         * or directly using SR->entry.base address.
                         * since virtual == physical addresses.
                         */
                        freeFramePoolobj->frmBufsStartAddrp = (UInt32)
                                SyslinkMemMgr_translate (
                                        obj->memMgrAllocater_frmBufs,
                                        (Ptr)(lfrmChunkStartAddr),
                                        SyslinkMemMgr_AddrType_Virtual,
                                        SyslinkMemMgr_AddrType_Portable);

                        freeFramePoolobj->frmBufsEndAddrp = (UInt32)
                                SyslinkMemMgr_translate (
                                        obj->memMgrAllocater_frmBufs,
                                        (Ptr)( lfrmChunkStartAddr
                                                +(numframes *tframbufSize)),
                                        SyslinkMemMgr_AddrType_Virtual,
                                        SyslinkMemMgr_AddrType_Portable);
                    }
                    frmbufOffset = 0;

                    /* Now prepare frame header for the frames in the pool*/
                    for (j = 0; j < freeFramePoolobj->numFrms; j++) {
                        frame = (FrameQBufMgr_ShMem_Frame)(lhdrChunkStartAddr
                                     +( (freeFramePoolobj->frmHdrBufSize) * j));

                        /*Size of the frame  header(total size of base header
                         *and extended header .
                         */
                        frame->headerSize =
                                         freeFramePoolobj->frmHdrBufSize;
                        /* FrameQBufMgr Id from which this frame is allocated */
                        frame->frmAllocaterId = obj->instId;
                        /* Free ququeNo from which frame is allocated in
                         * FrameQBufMgr
                         */
                        frame->freeFrmQueueNo = i;
                        /* Number of framebuffers  in frame. */
                        frame->numFrameBuffers = numBufsInFrm;
                        /*Size of the base Frame header*/
                        frame->baseHeaderSize =
                            FrameQBufMgr_getBaseHeaderSize(frame->numFrameBuffers);

                        frameBufInfo = &frame->frameBufInfo[0];

                        for (k = 0; k < numBufsInFrm;k++) {

                            /*frame buffer  in portable format*/
                            if (obj->bufInterfaceType
                                == FrameQBufMgr_ShMem_BUFINTERFACE_SHAREDMEM){

                                /*
                                 * On the RTOS this translation will actually
                                 * work whether allocation was from SR Heap
                                 * or directly using SR->entry.base address.
                                 * since virtual == physical addresses.
                                 */
                                frameBufInfo[k].bufPtr = (UInt32)
                                    SyslinkMemMgr_translate (
                                        obj->memMgrAllocater_frmBufs,
                                       (Ptr)(lfrmChunkStartAddr + frmbufOffset),
                                            SyslinkMemMgr_AddrType_Virtual,
                                            SyslinkMemMgr_AddrType_Portable);
                                /*frame buffer size */
                                frameBufInfo[k].bufSize =
                                         freeFramePoolobj->frmBufSize[k];
                                frmbufOffset +=
                                        freeFramePoolobj->frmBufSize[k];
                                /* valid data start offset from buffer start
                                 * addr */
                                frameBufInfo[k].startOffset= 0;
                                /*frame buffer valid size  from the startoffser.
                                 *startoffset+validsize must be less than or i
                                 * equal to bufSize
                                 */
                                frameBufInfo[k].validSize = 0;
                            }
                            if(obj->bufInterfaceType == FrameQBufMgr_ShMem_INTERFACE_TILERMEM){
                                frameBufInfo[k].pixelFormat = allocParams.block[0].pixelFormat     = params->frameBufParams[i][k].pixelFormat;
                                frameBufInfo[k].width       = allocParams.block[0].dim.area.width  = params->frameBufParams[i][k].width;
                                frameBufInfo[k].height      = allocParams.block[0].dim.area.height = params->frameBufParams[i][k].height;
                                allocParams.tilerBlocks     = 1;

                                frameBufInfo[k].bufPtr = (UInt32)
                                                         SyslinkMemMgr_alloc(obj->memMgrAllocater_frmBufs,
                                                                             &allocParams);
                                /* Return the stride to use*/
                                frameBufInfo[k].stride = allocParams.block[0].stride;
                                /* valid data start offset from buffer start addr */
                                frameBufInfo[k].startOffset= 0;
                            }
                        }
                        /*Write back the contents of this frame to memory*/
                        if (obj->frmHdrBufCacheFlag == TRUE) {
                            Cache_wbInv((Ptr)frame,
                                        frame->baseHeaderSize,
                                        Cache_Type_ALL,
                                        TRUE);
                        }
                    }
                    /* Increment  for the next free frame pool preparation*/
                    lhdrChunkStartAddr += ((numframes + numHdrs) * hdrSize);
                    lfrmChunkStartAddr += (numframes *tframbufSize);

                    if (obj->ctrlStructCacheFlag == TRUE) {
                        Cache_wbInv((Ptr)freeFramePoolobj,
                                    sizeof(FrameQBufMgr_ShMem_FreeFramePoolObj),
                                    Cache_Type_ALL,
                                    TRUE);
                    }

                }

            }
            else {
                status = FrameQBufMgr_E_FAIL;
                Error_raise(eb,
                            FrameQBufMgr_ShMem_E_Fail,
                            NULL,
                            0);
            }
        }

        if (status >= 0) {
           /* Last thing, set the status */
            pAttrs->status = FrameQBufMgr_ShMem_CREATED;
            if (obj->ctrlStructCacheFlag == TRUE) {
                Cache_wbInv((Ptr)pAttrs,
                            sizeof(FrameQBufMgr_ShMem_Attrs),
                            Cache_Type_ALL,
                            TRUE);
            }
        }
    }

    return (status);
}

UInt32 ti_syslink_ipc_rtos_shMem_FrameQBufMgr_ShMem_sharedMemReq__F(const ti_syslink_ipc_rtos_shMem_FrameQBufMgr_ShMem_Params *params,
                                UInt32                    *frmHdrBufChunkSize,
                                UInt32                    *frmBufChunkSize)
{
    FrameQBufMgr_ShMem_Params shparams;
    /*TBD: Copy shared params from RTSC params */
    FrameQBufMgr_ShMem_getSharedParams(&shparams, params);
    return FrameQBufMgr_ShMem_sharedMemReq(&shparams,
                                           frmHdrBufChunkSize,
                                           frmBufChunkSize);
}
